
import numpy as np
import json
from json import JSONEncoder

from pyflowline.classes.vertex import pyvertex
from pyflowline.classes.edge import pyedge
from pyflowline.classes.cell import pycell
from pyflowline.classes.flowline import pyflowline

from pyflowline.algorithms.auxiliary.gdal_functions import calculate_polygon_area

class MpasClassEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.float32):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, list):
            pass  
        if isinstance(obj, pyvertex):
            return json.loads(obj.tojson()) #lVertexID
        if isinstance(obj, pyedge):
            return obj.lEdgeID        
        if isinstance(obj, pyflowline):
            return obj.lFlowlineID
        
        if isinstance(obj, pympas):
            return obj.lCellID
            
        return JSONEncoder.default(self, obj)



class pympas(pycell):
    lCellID  = -1    
    nFlowline=0
    nVertex =0 
    nEdge=0
    dLength=0.0
    dArea=0.0
    dX_center_meter=0.0
    dY_center_meter=0.0
    dz_center=0.0
    dLongitude_center_degree=0.0
    dLatitude_center_degree=0.0
    dElevation_mean=0.0
    dElevation_profile0=0.0
    dLength_flowline=0.0
    iFlag_intersected=-1
    iFlag_coast = 0
    lCellID_downstream_burned=-1
    iStream_order_burned=-1
    iStream_segment_burned=-1
    aEdge=None
    aEdgeID=None
    aVertex=None
    aVertexID=None
    pVertex_center = None
    aFlowline=None    
    nNeighbor=-1
    nNeighbor_land=-1
    nNeighbor_ocean=-1
    aNeighbor=None #the global ID of all neighbors
    aNeighbor_land=None #the global ID of all neighbors
    aNeighbor_ocean=None #the global ID of all neighbors
    aNeighbor_distance = None

    def __init__(self, dLon, dLat, aEdge, aVertex):    
        nEdge = len(aEdge)
        if nEdge < 3 or nEdge > 8:
            print('At lease 3 edges are required!')
            pass
        else:                          
            self.aEdge = aEdge
            self.aVertex = aVertex #the first one and last one are the same
            self.nEdge = len(aEdge)
            self.nVertex = len(aVertex) 
            self.nNeighbor = -1
            self.nNeighbor_land = -1
            self.nNeighbor_ocean = -1
            self.iFlag_coast = 0      
            self.dLongitude_center_degree = dLon
            self.dLatitude_center_degree = dLat
            pVertex = dict()        
            pVertex['dLongitude_degree'] =self.dLongitude_center_degree
            pVertex['dLatitude_degree'] =self.dLatitude_center_degree           
            self.pVertex_center = pyvertex(pVertex)
            self.lCellID_downstream_burned=-1
            self.iStream_order_burned=-1
            self.iStream_segment_burned=-1
            self.dElevation_mean=-9999.0
            self.dElevation_profile0=-9999.0
            pass
        pass
    
   
    def has_this_edge(self, pEdge_in):
        iFlag_found = 0
        for pEdge in self.aEdge:
            if pEdge.is_overlap(pEdge_in):
                iFlag_found =1 
                break
            else:
                pass       
        
        return iFlag_found

    def which_edge_cross_this_vertex(self, pVertex_in):
        iFlag_found = 0
        pEdge_out = None
        for pEdge in self.aEdge:
            iFlag, dummy ,diff = pEdge.check_vertex_on_edge(pVertex_in)
            if( iFlag ==1 ):
                iFlag_found =1
                pEdge_out = pEdge
                break
            else:
                pass

        return iFlag_found, pEdge_out
    
    def calculate_cell_area(self): #not used anymore
        lons=list()
        lats=list()        
        for i in range(self.nVertex):            
            lons.append( self.aVertex[i].dLongitude_degree )
            lats.append( self.aVertex[i].dLatitude_degree )

        self.dArea = calculate_polygon_area(lons,lats )
        return self.dArea

    def calculate_edge_length(self):        
        self.dLength_edge = np.sqrt( self.dArea )
        return self.dLength_edge
    
    def share_edge(self, other):
        iFlag_share = 0
        for pEdge in self.aEdge:
            for pEdge2 in other.aEdge:
                if pEdge.is_overlap(pEdge2) ==1 :
                    iFlag_share = 1 
                    break

        return iFlag_share

    
    def tojson(self):
        aSkip = ['aEdge', \
                'aFlowline']
        obj = self.__dict__.copy()
        for sKey in aSkip:
            obj.pop(sKey, None)
        sJson = json.dumps(obj, \
            sort_keys=True, \
            indent = 4, \
            ensure_ascii=True, \
            cls=MpasClassEncoder)
        return sJson



import numpy as np
def remove_flowline_loop(aFlowline_in):
    aFlowline_out=list()
    nFlowline = len(aFlowline_in)
    def find_paralle_stream( pVertex_start_in):        
        ndownstream=0
        aDownstream=list()
        aStream_order_out=list()
        for j in range(nFlowline):
            pFlowline = aFlowline_in[j]
            pVertex_start = pFlowline.pVertex_start
            pVertex_end = pFlowline.pVertex_end
            if pVertex_start == pVertex_start_in: 
                ndownstream= ndownstream+1
                aDownstream.append(j)
                aStream_order_out.append(  pFlowline.iStream_order  )
                pass                
        return ndownstream, aDownstream, aStream_order_out

    lID=0
    aFlag = np.full(nFlowline, 0, dtype=int)
    for i in range(nFlowline):
        pFlowline = aFlowline_in[i]      
        pVertex_start = pFlowline.pVertex_start
        pVertex_end = pFlowline.pVertex_end
        iStream_order = pFlowline.iStream_order        
        ndownstream , aDownstream, aStream_order = find_paralle_stream( pVertex_start)
        if ndownstream == 1:
            if aFlag[i] !=1:
                pFlowline.lIndex = lID
                aFlowline_out.append(pFlowline)
                lID = lID + 1
                aFlag[i]=1
            pass
        else:                     
            #more than one, so we only take the current one or high order one            
            if(ndownstream>1):    
                sort_index = np.argsort(aStream_order)            
                sort_index = sort_index[::-1]
                lIndex = aDownstream[ sort_index[0] ]
                pFlowline_down =  aFlowline_in[lIndex]          
                if aFlag[ lIndex ]  !=1:
                    pFlowline_down.lIndex = lID
                    aFlowline_out.append(pFlowline_down)
                    lID = lID + 1
                    aFlag[lIndex]=1
                    pass
                
                for k in range(ndownstream):
                    aFlag[ aDownstream[k] ] = 1
            pass        
        pass 

    return aFlowline_out


#create a rectangle latitude/longitude based mesh
#we will use some GIS way to define it
#longitude left and latitude bottom and nrow and ncolumn and resolution is used to define the rectangle
#because it is mesh, it represent the edge instead of center
#we will use gdal api for most operations
import os, sys
import numpy as np
from osgeo import ogr, osr
from pyflowline.classes.hexagon import pyhexagon
from pyflowline.formats.convert_coordinates import convert_gcs_coordinates_to_cell, convert_pcs_coordinates_to_cell
from pyflowline.algorithms.auxiliary.find_index_in_list import check_if_duplicates
from pyflowline.algorithms.auxiliary.gdal_functions import reproject_coordinates_batch

def create_hexagon_mesh(iFlag_rotation_in, \
    dX_left_in, dY_bot_in, \
    dResolution_meter_in, \
        ncolumn_in, \
            nrow_in, \
        sFilename_output_in, \
            sFilename_spatial_reference_in):
    
    if os.path.exists(sFilename_output_in): 
        os.remove(sFilename_output_in)
    
    if os.path.exists(sFilename_spatial_reference_in): 
        pass
    else:
        print('Spatial reference file is missing')
        return

    pDriver_shapefile = ogr.GetDriverByName('Esri Shapefile')
    pDriver_geojson = ogr.GetDriverByName('GeoJSON')
    pDataset_shapefile = pDriver_shapefile.Open(sFilename_spatial_reference_in, 0)
    pLayer_shapefile = pDataset_shapefile.GetLayer(0)
    pSpatial_reference = pLayer_shapefile.GetSpatialRef()    
    pDataset = pDriver_geojson.CreateDataSource(sFilename_output_in)        
    pSpatial_reference_gcs = osr.SpatialReference()  
    pSpatial_reference_gcs.ImportFromEPSG(4326)    # WGS84 lat/lon
    pSpatial_reference_gcs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    pLayer = pDataset.CreateLayer('cell', pSpatial_reference_gcs, ogr.wkbPolygon)
    # Add one attribute
    pLayer.CreateField(ogr.FieldDefn('id', ogr.OFTInteger64)) #long type for high resolution
    pLayer.CreateField(ogr.FieldDefn('lon', ogr.OFTReal)) #long type for high resolution
    pLayer.CreateField(ogr.FieldDefn('lat', ogr.OFTReal)) #long type for high resolution
    pArea_field = ogr.FieldDefn('area', ogr.OFTReal)
    pArea_field.SetWidth(20)
    pArea_field.SetPrecision(2)
    pLayer.CreateField(pArea_field)    
    pLayerDefn = pLayer.GetLayerDefn()
    pFeature = ogr.Feature(pLayerDefn)
    xleft = dX_left_in
    ybottom = dY_bot_in
    dArea = np.power(dResolution_meter_in,2.0)
    #hexagon edge
    dLength_edge = np.sqrt(  2.0 * dArea / (3.0* np.sqrt(3.0))  )
    #geojson
    aHexagon=list()
    #.........
    #(x2,y2)-----(x3,y3)
    #   |           |
    #(x1,y1)-----(x4,y4)
    #...............
  
    lCellID = 1
    if iFlag_rotation_in ==0:
        dX_shift = 0.5 * dLength_edge * np.sqrt(3.0)
        dY_shift = 0.5 * dLength_edge
        dX_spacing = dLength_edge * np.sqrt(3.0)
        dY_spacing = dLength_edge * 1.5
        for iRow in range(1, nrow_in+1):
            for iColumn in range(1, ncolumn_in+1):
                if iRow % 2 == 1 : #odd
                #define a polygon here
                    x1 = xleft + (iColumn-1) * dX_spacing
                    y1 = ybottom + (iRow -1) * dY_spacing
                else:
                    x1 = xleft + (iColumn-1) * dX_spacing + dX_shift
                    y1 = ybottom + (iRow -1) * dY_spacing    
    
                x2 = x1 
                y2 = y1 + dLength_edge
    
                x3 = x1 + dX_shift
                y3 = y2 + dY_shift
    
                x4 = x1 + dX_spacing
                y4 = y2
    
                x5 = x4 
                y5 = y1 
    
                x6 = x3
                y6 = y1  -   dY_shift                    
                aCoords = np.full((7,2), -9999.0, dtype=float)
                
                x = list()
                x.append(x1)
                x.append(x2)
                x.append(x3)
                x.append(x4)
                x.append(x5)
                x.append(x6)

                y = list()
                y.append(y1)
                y.append(y2)
                y.append(y3)
                y.append(y4)
                y.append(y5)
                y.append(y6)

                x_new , y_new = reproject_coordinates_batch(x, y, pSpatial_reference, \
                    spatial_reference_target = pSpatial_reference_gcs)
                x1=x_new[0]
                x2=x_new[1]
                x3=x_new[2]
                x4=x_new[3]
                x5=x_new[4]
                x6=x_new[5]

                y1=y_new[0]
                y2=y_new[1]
                y3=y_new[2]
                y4=y_new[3]
                y5=y_new[4]
                y6=y_new[5]
    
                ring = ogr.Geometry(ogr.wkbLinearRing)
                ring.AddPoint(x1, y1)
                ring.AddPoint(x2, y2)
                ring.AddPoint(x3, y3)
                ring.AddPoint(x4, y4)
                ring.AddPoint(x5, y5)
                ring.AddPoint(x6, y6)
                ring.AddPoint(x1, y1)
                
                pPolygon = ogr.Geometry(ogr.wkbPolygon)
                pPolygon.AddGeometry(ring)    
                pFeature.SetGeometry(pPolygon)
                pFeature.SetField("id", lCellID)    
                
                aCoords[0,0] = x1
                aCoords[0,1] = y1
                aCoords[1,0] = x2
                aCoords[1,1] = y2
                aCoords[2,0] = x3
                aCoords[2,1] = y3
                aCoords[3,0] = x4
                aCoords[3,1] = y4
                aCoords[4,0] = x5
                aCoords[4,1] = y5
                aCoords[5,0] = x6
                aCoords[5,1] = y6
                aCoords[6,0] = x1
                aCoords[6,1] = y1
           
                dummy1= np.array(aCoords)
                dLongitude_center = np.mean(aCoords[0:6,0])
                dLatitude_center = np.mean(aCoords[0:6,1])
                pFeature.SetField("lon", dLongitude_center )
                pFeature.SetField("lat", dLatitude_center )
                pFeature.SetField("area", dArea )
                pLayer.CreateFeature(pFeature)
                pHexagon = convert_gcs_coordinates_to_cell(1, dLongitude_center, dLatitude_center, dummy1)
                pHexagon.lCellID = lCellID
                dArea = pHexagon.calculate_cell_area()
                pHexagon.dArea = dArea
                pHexagon.calculate_edge_length()
                pHexagon.dLongitude_center_degree = dLongitude_center
                pHexagon.dLatitude_center_degree = dLatitude_center

                lCellID_center = lCellID
                #build topoloy
                aNeighbor=list()
          
                if iColumn > 1:#0
                    lCellID0 = lCellID_center - 1
                    aNeighbor.append(lCellID0)

                if iRow < nrow_in :#1 and 2
                    if iRow %2 ==0:
                        lCellID1 = ncolumn_in * iRow + iColumn 
                        aNeighbor.append(lCellID1)
                       
                        if iColumn!=ncolumn_in:
                            lCellID2 = ncolumn_in * iRow + iColumn + 1
                            aNeighbor.append(lCellID2)
                          
                    else:
                        lCellID2 = ncolumn_in * iRow + iColumn
                        aNeighbor.append(lCellID2)
                    
                        if iColumn != 1:
                            lCellID1 = ncolumn_in * iRow + iColumn - 1 
                            aNeighbor.append(lCellID1)
                          
                        
                if iColumn < ncolumn_in:#3
                    lCellID3 = lCellID_center + 1
                    aNeighbor.append(lCellID3)             

                if iRow > 1 : #4 and 5
                    if iRow %2 ==1:
                        lCellID4 = ncolumn_in * (iRow-2) + iColumn 
                        aNeighbor.append(lCellID4)
                        

                        if iColumn !=1:
                            lCellID5 = ncolumn_in * (iRow-2) + iColumn -1
                            aNeighbor.append(lCellID5)
                           
                    else:
                        lCellID5 = ncolumn_in * (iRow-2) + iColumn 
                        aNeighbor.append(lCellID5)
                     
                        if iColumn!=ncolumn_in:
                            lCellID4 = ncolumn_in * (iRow-2) + iColumn + 1
                            aNeighbor.append(lCellID4)
                            

                if check_if_duplicates(aNeighbor) == 0:
                    print('error')        

                pHexagon.aNeighbor = aNeighbor
                pHexagon.nNeighbor = len(aNeighbor)
                pHexagon.aNeighbor_land= aNeighbor
                pHexagon.nNeighbor_land= pHexagon.nNeighbor
                

                aHexagon.append(pHexagon)
                lCellID= lCellID + 1    
                pass
              
        pass
    else:
        dX_shift = 0.5 * dLength_edge
        dY_shift = 0.5 * dLength_edge * np.sqrt(3.0)
        dX_spacing = dLength_edge * 1.5
        dY_spacing = dLength_edge * np.sqrt(3.0)
        for iColumn in range(1, ncolumn_in+1):
            for iRow in range(1, nrow_in+1):
                if iColumn % 2 == 0 :
                #define a polygon here
                    x1 = xleft + (iColumn-1) * dX_spacing
                    y1 = ybottom + (iRow-2) * dY_spacing + dY_shift
                else:
                    x1 = xleft + (iColumn-1) * dX_spacing #- dX_shift
                    y1 = ybottom + (iRow-1) * dY_spacing     
    
                x2 = x1 - dX_shift
                y2 = y1 + dY_shift
    
                x3 = x1 
                y3 = y1 + dY_shift * 2.0
    
                x4 = x1 + dLength_edge
                y4 = y1 + dY_shift * 2.0
    
                x5 = x4 + dX_shift
                y5 = y1 + dY_shift
    
                x6 = x1 + dLength_edge
                y6 = y1         
               
                aCoords = np.full((7,2), -9999.0, dtype=float)

                x = list()
                x.append(x1)
                x.append(x2)
                x.append(x3)
                x.append(x4)
                x.append(x5)
                x.append(x6)

                y = list()
                y.append(y1)
                y.append(y2)
                y.append(y3)
                y.append(y4)
                y.append(y5)
                y.append(y6)

                x_new , y_new = reproject_coordinates_batch(x, y, pSpatial_reference)
                x1=x_new[0]
                x2=x_new[1]
                x3=x_new[2]
                x4=x_new[3]
                x5=x_new[4]
                x6=x_new[5]

                y1=y_new[0]
                y2=y_new[1]
                y3=y_new[2]
                y4=y_new[3]
                y5=y_new[4]
                y6=y_new[5]
    
                ring = ogr.Geometry(ogr.wkbLinearRing)
                ring.AddPoint(x1, y1)
                ring.AddPoint(x2, y2)
                ring.AddPoint(x3, y3)
                ring.AddPoint(x4, y4)
                ring.AddPoint(x5, y5)
                ring.AddPoint(x6, y6)
                ring.AddPoint(x1, y1)
                pPolygon = ogr.Geometry(ogr.wkbPolygon)
                pPolygon.AddGeometry(ring)
    
                pFeature.SetGeometry(pPolygon)
                pFeature.SetField("id", lCellID)
    
                aCoords[0,0] = x1
                aCoords[0,1] = y1
                aCoords[1,0] = x2
                aCoords[1,1] = y2
                aCoords[2,0] = x3
                aCoords[2,1] = y3
                aCoords[3,0] = x4
                aCoords[3,1] = y4
                aCoords[4,0] = x5
                aCoords[4,1] = y5
                aCoords[5,0] = x6
                aCoords[5,1] = y6
                aCoords[6,0] = x1
                aCoords[6,1] = y1
                    
                dummy1= np.array(aCoords)
                dLongitude_center = np.mean(aCoords[0:6,0])
                dLatitude_center = np.mean(aCoords[0:6,1])       
                pHexagon = convert_gcs_coordinates_to_cell(1, dLongitude_center, dLatitude_center, dummy1)
                pHexagon.lCellID = lCellID
                #build topology
                dArea = pHexagon.calculate_cell_area()
                pHexagon.dArea = dArea
                pHexagon.calculate_edge_length()       
                pFeature.SetField("lon", dLongitude_center )
                pFeature.SetField("lat", dLatitude_center )
                pFeature.SetField("area", dArea )
                pLayer.CreateFeature(pFeature)
                lCellID_center = lCellID                
                aNeighbor=list()
                if iRow > 1:#0
                    lCellID0 = lCellID_center - 1
                    aNeighbor.append(lCellID0)

                if iColumn> 1:#1 ans 2
                    if iColumn %2 ==0:
                        lCellID1 = nrow_in * (iColumn-2) + iRow 
                        aNeighbor.append(lCellID1)
                        if iRow!=nrow_in:
                            lCellID2 = nrow_in * (iColumn-2) + iRow +1
                            aNeighbor.append(lCellID2)
                    else:
                        lCellID2 = nrow_in * (iColumn-2) + iRow
                        aNeighbor.append(lCellID2)
                        if iRow != 1:
                            lCellID1 = nrow_in * (iColumn-2) + iRow -1
                            aNeighbor.append(lCellID1)
                                        
                        
                if iRow < nrow_in:#3
                    lCellID3 = lCellID_center + 1
                    aNeighbor.append(lCellID3)

                if iColumn  < ncolumn_in  : #4 and 5
                    if iColumn %2 ==1:
                        lCellID4 = nrow_in * iColumn + iRow 
                        aNeighbor.append(lCellID4)
                        if iRow !=1:
                            lCellID5 = nrow_in * iColumn + iRow -1
                            aNeighbor.append(lCellID5)
                    else:
                        lCellID5 = nrow_in * iColumn + iRow 
                        aNeighbor.append(lCellID5)
                        if iRow!=nrow_in:
                            lCellID4 = nrow_in * iColumn + iRow  +1
                            aNeighbor.append(lCellID4)
                                                   
                if check_if_duplicates(aNeighbor) == 0:
                    print('error')  

                pHexagon.aNeighbor = aNeighbor
                pHexagon.nNeighbor = len(aNeighbor)
                pHexagon.aNeighbor_land= aNeighbor
                pHexagon.nNeighbor_land= pHexagon.nNeighbor
                aHexagon.append(pHexagon)

                lCellID= lCellID +1
    
                pass
       
        
    pDataset = pLayer = pFeature  = None      
    #calculate neighbor distance
    for pHexagon in aHexagon:
        aNeighbor = pHexagon.aNeighbor
        pHexagon.aNeighbor_distance=list()
        for lCellID1 in aNeighbor:
            for pHexagon1 in aHexagon:
                if pHexagon1.lCellID == lCellID1:
                    dDistance = pHexagon.pVertex_center.calculate_distance( pHexagon1.pVertex_center )
                    pHexagon.aNeighbor_distance.append(dDistance)
                    break

    return aHexagon






import sys
from pathlib import Path
from os.path import realpath
import argparse
import logging
for handler in logging.root.handlers[:]:
    logging.root.removeHandler(handler)

logging.basicConfig(format='%(asctime)s %(message)s')
logging.warning('is the time Pyflowline simulation started.')

from pyflowline.pyflowline_generate_template_configuration_file import pyflowline_generate_template_configuration_file

parser = argparse.ArgumentParser()
parser.add_argument("--sMesh_type", help = "sMesh_type",  type = str)
parser.add_argument("--iCase_index", help = "iCase_index",  type = int)
parser.add_argument("--dResolution_meter", help = "dResolution_meter",  type = float)
parser.add_argument("--sDate", help = "sDate",  type = str)

#example
#python notebook.py --sMesh_type hexagon --iCase_index 1 --dResolution_meter 50000 --sDate 20220201
pArgs = parser.parse_args()
if len(sys.argv) == 1:
    sMesh_type = 'mpas'
    iCase_index = 1
    dResolution_meter=5000
    sDate='20220308'
else:
    if len(sys.argv)> 1:
        sMesh_type = pArgs.sMesh_type
        iCase_index = pArgs.iCase_index
        dResolution_meter=pArgs.dResolution_meter
        sDate = pArgs.sDate
        print(sMesh_type, iCase_index, dResolution_meter, sDate)
    else:
        print(len(sys.argv), 'Missing arguments')
        pass

sPath = str( Path().resolve() )
iFlag_option = 1
sWorkspace_data = realpath( sPath +  '/data/susquehanna' )
sWorkspace_input =  str(Path(sWorkspace_data)  /  'input')
sWorkspace_output=  str(Path(sWorkspace_data)  /  'output')

sFilename_configuration_in = realpath( sPath +  '/tests/configurations/template.json' )
    
oPyflowline = pyflowline_generate_template_configuration_file(sFilename_configuration_in,\
         sWorkspace_input, sWorkspace_output, iFlag_use_mesh_dem_in = 1,sMesh_type_in=sMesh_type, iCase_index_in = iCase_index, sDate_in = sDate)
    
print(oPyflowline.tojson())

print('Finished')

logging.basicConfig(format='%(asctime)s %(message)s')
logging.warning('is the time Pyflowline simulation finished.')



import os
from osgeo import ogr, osr
from shapely.geometry import Point

def export_vertex_to_json(aVertex_in, \
        sFilename_json_in,\
        iFlag_projected_in=None,\
        pSpatial_reference_in=None, \
        aAttribute_data=None):
    """
    convert a shapefile to json format.
    This function should be used for stream flowline only.
    """

    if os.path.exists(sFilename_json_in): 
        os.remove(sFilename_json_in)
        pass
    
    if iFlag_projected_in is None:
        iFlag_projected_in = 0
    else:
        iFlag_projected_in = 1

    if  pSpatial_reference_in is None:        
        pSpatial_reference_in = osr.SpatialReference()  
        pSpatial_reference_in.ImportFromEPSG(4326)    # WGS84 lat/lon
    else:
        pass

    if aAttribute_data is not None:
        aAttribute= aAttribute_data
        iFlag_attribute =1
    else:
        iFlag_attribute=0

    nVertex = len(aVertex_in)
    pDriver = ogr.GetDriverByName('GeoJSON')        
    pDataset_json = pDriver.CreateDataSource(sFilename_json_in)
    pLayer_json = pDataset_json.CreateLayer('vertex', pSpatial_reference_in, ogr.wkbPoint)
    # Add one attribute
    pLayer_json.CreateField(ogr.FieldDefn('id', ogr.OFTInteger64)) #long type for high resolution
    if iFlag_attribute ==1:        
        pLayer_json.CreateField(ogr.FieldDefn('con', ogr.OFTInteger64)) #long type for high resolution
        pass

    pLayerDefn = pLayer_json.GetLayerDefn()
    pFeature_out = ogr.Feature(pLayerDefn)
    lID = 0
    for i in range(nVertex):       
        pVertex = aVertex_in[i]
        if iFlag_projected_in ==1:
            dummy1= Point( pVertex.dx, pVertex.dy )
            pass
        else:
            dummy1= Point( pVertex.dLongitude_degree, pVertex.dLatitude_degree ) 
            pass

        pGeometry_out = ogr.CreateGeometryFromWkb(dummy1.wkb)
        pFeature_out.SetGeometry(pGeometry_out)   
        pFeature_out.SetField("id", lID)
        if iFlag_attribute ==1:
            pFeature_out.SetField("con", int(aAttribute[i]) )
                
        pLayer_json.CreateFeature(pFeature_out)        
        lID =  lID + 1
        pass
        
    pDataset_json.FlushCache()
    pDataset_json = pLayer_json = pFeature_out  = None    

    return


def export_vertex_to_shapefile(aVertex_in, sFilename_shapefile_out,\
    pSpatial_reference_in, 
    iFlag_projected_in, 
    aAttribute_data=None):
    """
    convert a shpefile to json format.
    This function should be used for stream flowline only.
    """

    if os.path.exists(sFilename_shapefile_out): 
        os.remove(sFilename_shapefile_out)
        pass

    if aAttribute_data is not None:
        aAttribute= aAttribute_data
        iFlag_attribute =1
    else:
        iFlag_attribute=0

    nVertex = len(aVertex_in)
    pDriver = ogr.GetDriverByName('ESRI Shapefile')
    pDataset_json = pDriver.CreateDataSource(sFilename_shapefile_out)
    pLayer_json = pDataset_json.CreateLayer('vertex', pSpatial_reference_in, ogr.wkbPoint)
    # Add one attribute
    pLayer_json.CreateField(ogr.FieldDefn('id', ogr.OFTInteger64)) #long type for high resolution
    if iFlag_attribute ==1:        
        pLayer_json.CreateField(ogr.FieldDefn('con', ogr.OFTInteger64)) #long type for high resolution
        pass

    pLayerDefn = pLayer_json.GetLayerDefn()
    pFeature_out = ogr.Feature(pLayerDefn)
    lID = 0
    for i in range(nVertex):       
        pVertex = aVertex_in[i]
        if iFlag_projected_in ==1:
            dummy1= Point( pVertex.dx, pVertex.dy )             
            pass
        else:
            dummy1= Point( pVertex.dLongitude_degree, pVertex.dLatitude_degree ) 
            pass

        pGeometry_out = ogr.CreateGeometryFromWkb(dummy1.wkb)
        pFeature_out.SetGeometry(pGeometry_out)   
        pFeature_out.SetField("id", lID)
        if iFlag_attribute ==1:
            pFeature_out.SetField("con", int(aAttribute[i]) )
        
        # Add new pFeature_shapefile to output Layer
        pLayer_json.CreateFeature(pFeature_out)        
        lID =  lID + 1
        pass
        
    pDataset_json.FlushCache()
    pDataset_json = pLayer_json = pFeature_out  = None    

    return


    


    



def check_head_water(aFlowline_in, pVertex_start_in):
    """[Check whether a vertex assoacited with a flowline is a headwater or not]

    Args:
        aFlowline_in ([pyflowline]): [all the flowline]
        pVertex_start_in ([pyvertex]): [the vertex of interest]

    Returns:
        [int]: [0: not headwater; 1: is headwater]
    """
    nFlowline = len(aFlowline_in)
    iFlag_head_water = -1
    iCount = 0
    for i in range(nFlowline):
        pFlowline = aFlowline_in[i]
        pVerter_start = pFlowline.pVertex_start
        pVerter_end = pFlowline.pVertex_end
        if pVerter_start == pVertex_start_in:
            iCount = iCount + 1
            pass
        if  pVerter_end == pVertex_start_in:
            iCount = iCount + 1
            pass
        pass
    if iCount == 1:
        iFlag_head_water=1
        
    return iFlag_head_water


import os
import numpy as np
from osgeo import ogr, gdal
from shapely.wkt import loads

from pyflowline.formats.convert_coordinates import convert_gcs_coordinates_to_cell

def read_mesh_json(iMesh_type_in, sFilename_mesh_in):
    """
    convert a shpefile to json format.
    This function should be used for stream flowline only.
    """
    iReturn_code = 1
    if os.path.isfile(sFilename_mesh_in):
        pass
    else:
        print('This mesh file does not exist: ', sFilename_mesh_in )
        iReturn_code = 0
        return iReturn_code

    aCell_out=list()
    pDriver_json = ogr.GetDriverByName('GeoJSON')    
    pDataset_mesh = pDriver_json.Open(sFilename_mesh_in, gdal.GA_ReadOnly)
    pLayer_mesh = pDataset_mesh.GetLayer(0)
    pSpatial_reference_out = pLayer_mesh.GetSpatialRef()
    ldefn = pLayer_mesh.GetLayerDefn()
    schema =list()
    for n in range(ldefn.GetFieldCount()):
        fdefn = ldefn.GetFieldDefn(n)
        schema.append(fdefn.name)
    if 'iseg' in schema:
        iFlag_segment = 1
    else:
        iFlag_segment = 0   

    #we also need to spatial reference
    for pFeature_mesh in pLayer_mesh:
        pGeometry_mesh = pFeature_mesh.GetGeometryRef()        
        dummy0 = loads( pGeometry_mesh.ExportToWkt() )
        aCoords_gcs = dummy0.exterior.coords
        aCoords_gcs= np.array(aCoords_gcs)       
        lCellID = pFeature_mesh.GetField("id")
        dLon = pFeature_mesh.GetField("lon")
        dLat = pFeature_mesh.GetField("lat")        
        dArea = pFeature_mesh.GetField("area")
        if iMesh_type_in == 4:
            dElevation_mean = pFeature_mesh.GetField("elev")
            dElevation_profile0 = pFeature_mesh.GetField("elev0")
    
        pGeometrytype_mesh = pGeometry_mesh.GetGeometryName()
        if(pGeometrytype_mesh == 'POLYGON'):            
            pCell = convert_gcs_coordinates_to_cell(iMesh_type_in, dLon, dLat, aCoords_gcs)     
            pCell.lCellID = lCellID          
            pCell.dArea = dArea 
            pCell.dLength = pCell.calculate_edge_length()
            pCell.dLength_flowline = pCell.dLength
            if iMesh_type_in == 4:
                pCell.dElevation_mean = dElevation_mean
                pCell.dElevation_profile0 = dElevation_profile0

            aCell_out.append(pCell)

    return aCell_out, pSpatial_reference_out




import copy
import numpy as np

def find_vertex_in_list(aVertex_in, pVertex_in):
    """[find the index of a vertex in a list]

    Args:
        aVertex_in ([type]): [description]
        pVertex_in ([type]): [description]

    Returns:
        [type]: [description]
    """

    iFlag_exist = 0
    lIndex= -1
    nVertex= len(aVertex_in)

    if nVertex > 0 :
        for i in np.arange( nVertex):
            pVertex = aVertex_in[i]
            if pVertex == pVertex_in:
                iFlag_exist = 1      
                lIndex = i 
                break                
            else:
                pass

        pass        
        
    else:
        pass
    
    return iFlag_exist, lIndex



def find_vertex_on_edge(aVertex_in, pEdge_in):
    iFlag_exist = 0
    aIndex= list()
    aIndex_order=list()
    aDistance=list()
    nVertex= len(aVertex_in)
    npoint = 0    
    if nVertex > 0 :
        for i in np.arange( nVertex):
            pVertex = aVertex_in[i]
            iFlag_overlap, dDistance, diff = pEdge_in.check_vertex_on_edge(pVertex)
            if iFlag_overlap == 1:                
                iFlag_exist = 1                      
                aDistance.append(dDistance)
                aIndex.append(i) 
                npoint = npoint + 1          
            else:                
                if  diff < 1.0:
                    iFlag_overlap = pEdge_in.check_vertex_on_edge(pVertex)
                   
                pass

        #re-order 
        if iFlag_exist == 1 :
            x = np.array(aDistance)
            b = np.argsort(x)
            c = np.array(aIndex)
            d= c[b]
            aIndex_order = list(d)        
    else:
        pass
    
    return iFlag_exist, npoint , aIndex_order


def find_edge_in_list(aEdge_in, pEdge_in):
    """[find the index of an edge in a list]

    Args:
        aEdge_in ([type]): [description]
        pEdge_in ([type]): [description]

    Returns:
        [type]: [description]
    """
    iFlag_exist = 0
    lIndex= -1
    nEdge= len(aEdge_in)

    if nEdge > 0 :
        for i in np.arange( nEdge):
            pEdge = aEdge_in[i]
            if pEdge == pEdge_in:
                iFlag_exist = 1      
                lIndex = i 
                break                
            else:
                pass

        pass        
        
    else:
        pass
    
    return iFlag_exist, lIndex


def find_flowline_in_list(aFlowline_in, pFlowline_in):
    """[find the index of a flowline in a list]

    Args:
        aFlowline_in ([type]): [description]
        pFlowline_in ([type]): [description]

    Returns:
        [type]: [description]
    """
    iFlag_exist = 0
    lIndex= -1
    nFlowline= len(aFlowline_in)

    if nFlowline > 0 :
        for i in np.arange( nFlowline):
            pFlowline = aFlowline_in[i]
            if pFlowline == pFlowline_in:
                iFlag_exist = 1      
                lIndex = i 
                break                
            else:
                pass

        pass        
        
    else:
        pass
    
    return iFlag_exist, lIndex

def find_hexagon_through_edge(aHexagon_in, pEdge_in):
    """find the hexagons which contain an edge

    Args:
        aHexagon_in ([type]): [description]
        pEdge_in ([type]): [description]

    Returns:
        [type]: [description]
    """

    nHexagon = len(aHexagon_in)
    aHexagon_out = list()
    for i in range(nHexagon):
        pHexagon = aHexagon_in[i]
        if pHexagon.has_this_edge(pEdge_in) ==1:
            aHexagon_out.append(pHexagon)
            pass
        else:
            pass

    return aHexagon_out

def check_if_duplicates(aList_in):
    """[Check if given list contains any duplicates]

    Returns:
        [type]: [description]
    """
    iFlag_unique = 1
    for elem in aList_in:
        if aList_in.count(elem) > 1:
            iFlag_unique = 0
            break
        else:
            pass
    
    return iFlag_unique


def add_unique_vertex(aVertex_in, pVertex_in):
    """[add a vertex to a list if it is not already included]

    Args:
        aVertex_in ([type]): [description]
        pVertex_in ([type]): [description]

    Returns:
        [type]: [description]
    """
    iFlag_exist = 0
    nVertex = len(aVertex_in)     

    iFlag_exist, dummy =  find_vertex_in_list(aVertex_in, pVertex_in)

    if iFlag_exist == 1:
        pass
    else:
        aVertex_in.append(pVertex_in)
        pass

    return aVertex_in, iFlag_exist


def find_list_in_list(aList_in, pList_in):
    c = copy.deepcopy(pList_in)
    c.sort()
    nList = len(aList_in)
    iFlag = 0
    for i in range(nList):
        a = aList_in[i]
        b = copy.deepcopy(a)
        b.sort()
        if (b == c ):
            iFlag =1 
            break
        else:
            iFlag = 0
   
    return iFlag



from pyflowline.algorithms.auxiliary.check_head_water import check_head_water
def remove_small_river(aFlowline_in, dThreshold_in):
    nFlowline = len(aFlowline_in)
    aFlowline_out=list()        
    if nFlowline == 1:
        aFlowline_out.append(aFlowline_in[0])
    else:
        lID = 0        
        for i in range(nFlowline):
            pFlowline = aFlowline_in[i]      
            iFlag_dam = pFlowline.iFlag_dam
            pVertex_start = pFlowline.pVertex_start
            pVertex_end = pFlowline.pVertex_end
            dLength = pFlowline.calculate_length()
            if iFlag_dam ==1:
                pFlowline.lIndex = lID
                aFlowline_out.append(pFlowline)
                lID = lID +1       
            else:

                if check_head_water(aFlowline_in, pVertex_start)==1:
                    if dLength > dThreshold_in :
                        pFlowline.lIndex = lID
                        aFlowline_out.append(pFlowline)
                        lID = lID + 1 
                        pass
                    else:
                        pass
                else:        
                    pFlowline.lIndex = lID
                    aFlowline_out.append(pFlowline)
                    lID = lID +1       
                    pass            
            
            pass    

    return aFlowline_out


#create a rectangle latitude/longitude based mesh
#we will use some GIS way to define it
#longitude left and latitude bottom and nrow and ncolumn and resolution is used to define the rectangle
#because it is mesh, it represent the edge instead of center
#we will use gdal api for most operations
import os, sys
import numpy as np
from osgeo import ogr, osr
from pyflowline.classes.latlon import pylatlon
from pyflowline.formats.convert_coordinates import convert_gcs_coordinates_to_cell

def create_latlon_mesh(dLongitude_left_in, dLatitude_bot_in, dResolution_degree_in, ncolumn_in, nrow_in, sFilename_output_in):   
    if os.path.exists(sFilename_output_in): 
        os.remove(sFilename_output_in)

    #pDriver_shapefile = ogr.GetDriverByName('Esri Shapefile')
    pDriver_geojson = ogr.GetDriverByName('GeoJSON')    
    pDataset = pDriver_geojson.CreateDataSource(sFilename_output_in)
    pSpatial_reference_gcs = osr.SpatialReference()  
    pSpatial_reference_gcs.ImportFromEPSG(4326)    # WGS84 lat/lon
    pSpatial_reference_gcs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    pLayer = pDataset.CreateLayer('cell', pSpatial_reference_gcs, ogr.wkbPolygon)
    # Add one attribute
    pLayer.CreateField(ogr.FieldDefn('id', ogr.OFTInteger64)) #long type for high resolution
    pLayer.CreateField(ogr.FieldDefn('lon', ogr.OFTReal)) #long type for high resolution
    pLayer.CreateField(ogr.FieldDefn('lat', ogr.OFTReal)) #long type for high resolution
    pArea_field = ogr.FieldDefn('area', ogr.OFTReal)
    pArea_field.SetWidth(20)
    pArea_field.SetPrecision(2)
    pLayer.CreateField(pArea_field)    
    pLayerDefn = pLayer.GetLayerDefn()
    pFeature = ogr.Feature(pLayerDefn)    
    xleft = dLongitude_left_in
    xspacing= dResolution_degree_in
    ybottom = dLatitude_bot_in
    yspacing = dResolution_degree_in
    lCellID = 1
    #.........
    #(x2,y2)-----(x3,y3)
    #   |           |
    #(x1,y1)-----(x4,y4)
    #...............
    aLatlon = list()
    for iColumn in range(1, ncolumn_in+1):
        for iRow in range(1, nrow_in+1):
            #define a polygon here
            x1 = xleft + ((iColumn-1) * xspacing)
            y1 = ybottom + ((iRow-1) * yspacing)

            x2 = xleft + ((iColumn-1) * xspacing)
            y2 = ybottom + ((iRow ) * yspacing)

            x3 = xleft + ((iColumn ) * xspacing)
            y3 = ybottom + ((iRow ) * yspacing)

            x4 = xleft + ((iColumn ) * xspacing)
            y4 = ybottom + ((iRow-1) * yspacing)           

            ring = ogr.Geometry(ogr.wkbLinearRing)
            ring.AddPoint(x1, y1)
            ring.AddPoint(x2, y2)
            ring.AddPoint(x3, y3)
            ring.AddPoint(x4, y4)
            ring.AddPoint(x1, y1)
            pPolygon = ogr.Geometry(ogr.wkbPolygon)
            pPolygon.AddGeometry(ring)

            dLongitude_center = (x1 + x2 + x3 + x4)/4.0
            dLatitude_center = (y1 + y2 + y3 + y4)/4.0
            aCoords = np.full((5,2), -9999.0, dtype=float)
            aCoords[0,0] = x1
            aCoords[0,1] = y1
            aCoords[1,0] = x2
            aCoords[1,1] = y2
            aCoords[2,0] = x3
            aCoords[2,1] = y3
            aCoords[3,0] = x4
            aCoords[3,1] = y4
            aCoords[4,0] = x1
            aCoords[4,1] = y1
            dummy1= np.array(aCoords)           
            pLatlon = convert_gcs_coordinates_to_cell(3, dLongitude_center, dLatitude_center, dummy1)
            pLatlon.lCellID = lCellID
            dArea = pLatlon.calculate_cell_area()
            pLatlon.calculate_edge_length()
            pLatlon.dLongitude_center_degree = dLongitude_center
            pLatlon.dLatitude_center_degree = dLatitude_center
            pFeature.SetGeometry(pPolygon)
            pFeature.SetField("id", lCellID)
            pFeature.SetField("lon", dLongitude_center )
            pFeature.SetField("lat", dLatitude_center )
            pFeature.SetField("area", dArea )
            pLayer.CreateFeature(pFeature)

            lCellID_center = lCellID

            aNeighbor = list()
            if iRow > 1:#under
                lCellID0 = lCellID_center - 1
                aNeighbor.append(lCellID0)
                if iColumn > 1:
                    lCellID2 = lCellID0 - nrow_in
                    aNeighbor.append(lCellID2)

            if iColumn> 1:#left
                lCellID1 = nrow_in * (iColumn-2) + iRow 
                aNeighbor.append(lCellID1)  
                if iRow < nrow_in:
                    lCellID4 = lCellID1 + 1
                    aNeighbor.append(lCellID4)      
                    
            if iRow < nrow_in:#top
                lCellID3 = lCellID_center + 1
                aNeighbor.append(lCellID3)
                if iColumn < ncolumn_in:
                    lCellID6 = lCellID3 + nrow_in
                    aNeighbor.append(lCellID6) 
                    
            if iColumn  < ncolumn_in  : #right
                lCellID5 = nrow_in * iColumn + iRow 
                aNeighbor.append(lCellID5)
                if iRow > 1:
                    lCellID7 = lCellID5 -1
                    aNeighbor.append(lCellID7) 

            pLatlon.aNeighbor = aNeighbor
            pLatlon.nNeighbor = len(aNeighbor)
            pLatlon.aNeighbor_land= aNeighbor
            pLatlon.nNeighbor_land= pLatlon.nNeighbor            
            aLatlon.append(pLatlon)
            lCellID = lCellID + 1

            pass
        
    pDataset = pLayer = pFeature  = None      

    #calculate neighbor distance
    for pLatlon in aLatlon:
        aNeighbor = pLatlon.aNeighbor
        pLatlon.aNeighbor_distance=list()
        for lCellID1 in aNeighbor:
            for pLatlon1 in aLatlon:
                if pLatlon1.lCellID == lCellID1:
                    dDistance = pLatlon.pVertex_center.calculate_distance( pLatlon1.pVertex_center )
                    pLatlon.aNeighbor_distance.append(dDistance)
                    break

    return aLatlon








import os
import math
import numpy as np
from osgeo import ogr, osr
from netCDF4 import Dataset
from pyflowline.classes.mpas import pympas
from pyflowline.formats.convert_attributes import convert_gcs_attributes_to_cell
from pyflowline.algorithms.auxiliary.gdal_functions import convert_360_to_180

def create_mpas_mesh(iFlag_global_in, \
    iFlag_use_mesh_dem, \
        iFlag_save_mesh_in, \
        dLongitude_left_in, dLongitude_right_in,\
    dLatitude_top_in, dLatitude_bot_in, \
     sFilename_mesh_netcdf_in, \
         sFilename_output_in):
    
    if (os.path.exists(sFilename_mesh_netcdf_in)):
        pass
    else:
        print('Mesh file does not exist!')
        exit
    
    if os.path.exists(sFilename_output_in):  
        os.remove(sFilename_output_in)

    pDatasets_in = Dataset(sFilename_mesh_netcdf_in)

    netcdf_format = pDatasets_in.file_format
    pDriver_geojson = ogr.GetDriverByName('GeoJSON')     
    pSpatial_reference_gcs = osr.SpatialReference()  
    pSpatial_reference_gcs.ImportFromEPSG(4326)    # WGS84 lat/lon    
    #geojson
    if iFlag_save_mesh_in ==1:
        pDataset = pDriver_geojson.CreateDataSource(sFilename_output_in)
        pLayer = pDataset.CreateLayer('cell', pSpatial_reference_gcs, ogr.wkbPolygon)
        # Add one attribute
        pLayer.CreateField(ogr.FieldDefn('id', ogr.OFTInteger64)) #long type for high resolution
        pLayer.CreateField(ogr.FieldDefn('lon', ogr.OFTReal)) #long type for high resolution
        pLayer.CreateField(ogr.FieldDefn('lat', ogr.OFTReal)) #long type for high resolution
        pArea_field = ogr.FieldDefn('area', ogr.OFTReal)
        pArea_field.SetWidth(20)
        pArea_field.SetPrecision(2)
        pLayer.CreateField(pArea_field)
        if iFlag_use_mesh_dem == 1:
            pLayer.CreateField(ogr.FieldDefn('elev', ogr.OFTReal)) #float type for high resolution
            pLayer.CreateField(ogr.FieldDefn('elev0', ogr.OFTReal)) #float type for high resolution
        else:
            pass
    
        pLayerDefn = pLayer.GetLayerDefn()
        pFeature = ogr.Feature(pLayerDefn)        

    #write new netcdf
    for sKey, aValue in pDatasets_in.variables.items():      
        #we need to filter out unused grids based on mpas specs
        if sKey == 'latCell':
            latCell0 = aValue
        else:
            pass
        if sKey == 'lonCell':
            lonCell0 = aValue 
        else:
            pass
        
        if sKey == 'edgesOnCell':
            edgesOnCell0 = aValue 
        else:
            pass
            
        if sKey == 'cellsOnCell':
            cellsOnCell0 = aValue 
        else:
            pass

        if sKey == 'cellsOnEdge':
            cellsOnEdge0 = aValue 
        else:
            pass
            
        if sKey == 'verticesOnCell':
            verticesOnCell0 = aValue
        else:
            pass
        
        if sKey == 'verticesOnEdge':
            verticesOnEdge0 = aValue 
        else:
            pass

        if sKey == 'indexToCellID':
            indexToCellID0 = aValue 
        else:
            pass

        if sKey == 'indexToEdgeID':
            indexToEdgeID0 = aValue 
        else:
            pass

        if sKey == 'indexToVertexID':
            indexToVertexID0 = aValue 
        else:
            pass

        if sKey == 'lonVertex':
            lonVertex0 = aValue 
        else:
            pass

        if sKey == 'latVertex':
            latVertex0 = aValue 
        else:
            pass

        if sKey == 'areaCell':
            areaCell0 = aValue 
        else:
            pass

        if sKey == 'bed_elevation':
            bed_elevation0 = aValue 
        else:
            pass

        if sKey == 'ice_thickness':
            ice_thickness0 = aValue 
        else:
            pass

        if sKey == 'areaCell':
            areaCell0 = aValue 
        else:
            pass

        if sKey == 'dcEdge':
            dcEdge0 = aValue 
        else:
            pass

        if sKey == 'bed_elevation_profile':
            bed_elevation_profile0 = aValue 
        else:
            pass
        

    aLatitudeVertex = latVertex0[:] / math.pi * 180
    aLongitudeVertex = lonVertex0[:] / math.pi * 180    
    #convert unit 
    aLatitudeCell = latCell0[:] / math.pi * 180
    aLongitudeCell = lonCell0[:] / math.pi * 180
    aCellsOnCell = cellsOnCell0[:]
    aCellOnEdge = cellsOnEdge0[:]
    aEdgesOnCell = edgesOnCell0[:]
    aVertexOnCell = verticesOnCell0[:]
    aVertexOnEdge0 = verticesOnEdge0[:]    
    aIndexToCellID = indexToCellID0[:]
    aIndexToEdgeID = indexToEdgeID0[:]
    aIndexToVertexID = indexToVertexID0[:]
    aBed_elevation = bed_elevation0[:]
    aIce_thickness = ice_thickness0[:]
    aCellArea = areaCell0[:]
    aDcEdge = dcEdge0[:]
    aBed_elevation_profile = bed_elevation_profile0[:]  #elevation    
    ncell = len(aIndexToCellID) 
    aMpas = list()
    for i in range(ncell):
        dLat = convert_360_to_180 (aLatitudeCell[i])
        dLon = convert_360_to_180 (aLongitudeCell[i])
        if dLat > dLatitude_bot_in and dLat < dLatitude_top_in and dLon > dLongitude_left_in and dLon < dLongitude_right_in:
            #get cell edge
            lCellID = int(aIndexToCellID[i])
            dElevation_mean = float(aBed_elevation[i])
            dElevation_profile0 = float(aBed_elevation_profile[i,0])
            dThickness_ice = float( aIce_thickness[i] )
            dArea = float(aCellArea[i])
            if dThickness_ice > 0:
                continue
            else:
                aCellOnCellIndex = np.array(aCellsOnCell[i,:])
                aEdgesOnCellIndex = np.array(aEdgesOnCell[i,:])
                aVertexOnCellIndex = np.array(aVertexOnCell[i,:])
                dummy0 = np.where(aVertexOnCellIndex > 0)
                aVertexIndex = aVertexOnCellIndex[dummy0]
                dummy1 = np.where(aEdgesOnCellIndex > 0)
                aEdgeIndex= aEdgesOnCellIndex[dummy1]
                dummy2 = np.where(aCellOnCellIndex > 0)
                aNeighborIndex= (aCellOnCellIndex[dummy2]).astype(int)
                aVertexIndexOnEdge = np.array(aVertexOnEdge0[aEdgeIndex-1,:]).astype((int))
                aLonVertex = aLongitudeVertex[aVertexIndex-1]
                aLatVertex = aLatitudeVertex[aVertexIndex-1]
                nVertex = len(aLonVertex)
                ring = ogr.Geometry(ogr.wkbLinearRing)
                aCoords = np.full((nVertex,2), -9999.0, dtype=float)

                for j in range(nVertex):
                    x1 = convert_360_to_180(aLonVertex[j])
                    y1 = aLatVertex[j]      
                    ring.AddPoint(x1, y1)
                    aCoords[j,0] = x1
                    aCoords[j,1] = y1
                    pass

                if iFlag_save_mesh_in ==1:
                    x1 = convert_360_to_180(aLonVertex[0])
                    y1 = aLatVertex[0]
                    ring.AddPoint(x1, y1) #double check            
                    pPolygon = ogr.Geometry(ogr.wkbPolygon)
                    pPolygon.AddGeometry(ring)
                    pFeature.SetGeometry(pPolygon)
                    pFeature.SetField("id", int(lCellID) )
                    pFeature.SetField("lon", dLon )
                    pFeature.SetField("lat", dLat )
                    pFeature.SetField("area", dArea )
                    if iFlag_use_mesh_dem == 1:
                        pFeature.SetField("elev", dElevation_mean )
                        pFeature.SetField("elev0", dElevation_profile0 )

                    pLayer.CreateFeature(pFeature)

                pmpas = convert_gcs_attributes_to_cell(4, dLon, dLat, aCoords, aVertexIndex, aEdgeIndex, aVertexIndexOnEdge)               
                pmpas.dArea = dArea
                pmpas.calculate_edge_length()
                pmpas.dLength_flowline = pmpas.dLength_edge #Default
                pmpas.lCellID = lCellID
                pmpas.dElevation_mean  = dElevation_mean
                pmpas.dElevation_profile0 = dElevation_profile0
                pmpas.aNeighbor=aNeighborIndex
                pmpas.nNeighbor=len(aNeighborIndex)
                pmpas.aNeighbor_land=aNeighborIndex
                pmpas.nNeighbor_land=len(aNeighborIndex)
                aDistance=list()
                for i in range(pmpas.nNeighbor):
                    lNeighborID = pmpas.aNeighbor[i]
                    #find shared edge
                    lEdgeID= aEdgeIndex[i]                    
                    lIndex = aIndexToEdgeID[lEdgeID-1]
                    dDistance = aDcEdge[lIndex]
                    aDistance.append(dDistance)
                    pass

                pmpas.aNeighbor_distance = aDistance
                aMpas.append(pmpas)
                #get vertex

        pass

    #for maps we need to clean some cell because they were not actually in the domain

    if iFlag_global_in == 1:
        aMpas_out = aMpas
    else:
        aMpas_out = list()
        ncell = len(aMpas)
        aCellID  = list()
        for i in range(ncell):
            pCell = aMpas[i]
            lCellID = pCell.lCellID
            aCellID.append(lCellID)

        for i in range(ncell):
            pCell = aMpas[i]
            aNeighbor = pCell.aNeighbor
            nNeighbor = pCell.nNeighbor
            aNeighbor_new = list()
            nNeighbor_new = 0 
            for j in range(nNeighbor):
                lNeighbor = int(aNeighbor[j])
                if lNeighbor in aCellID:
                    nNeighbor_new = nNeighbor_new + 1 
                    aNeighbor_new.append(lNeighbor)

            pCell.nNeighbor_land= len(aNeighbor_new)
            pCell.aNeighbor_land = aNeighbor_new
            pCell.nNeighbor_ocean = pCell.nVertex - pCell.nNeighbor_land
            aMpas_out.append(pCell)


    return aMpas_out



import numpy as np 
from pyflowline.algorithms.auxiliary.check_head_water import check_head_water
lID=0
aFlag_process=None

def correct_flowline_direction(aFlowline_in, pVertex_outlet_in):        
    #we have to go reversely    
    aFlowline_out= list()   
    global lID    
    global aFlag_process
    nFlowline = len(aFlowline_in)
    aFlag_process=np.full(nFlowline, 0, dtype =int)
    iFlag_first = 1
    for i in range(nFlowline):
        pFlowline = aFlowline_in[i]
        iFlag_dam = pFlowline.iFlag_dam
        iStream_order = pFlowline.iStream_order
        pVertex_start = pFlowline.pVertex_start
        pVertex_end = pFlowline.pVertex_end
        dDiatance = pVertex_end.calculate_distance( pVertex_outlet_in )   
        if iFlag_first ==1:
            dDiatance_min = dDiatance                
            lIndex_outlet = i            
            iFlag_first=0    
        else:            
            if  dDiatance < dDiatance_min:
                dDiatance_min = dDiatance
                #found it
                lIndex_outlet = i      
                pass    
            else:
          
                pass
        
    lID = 0 
    pFlowline = aFlowline_in[lIndex_outlet]
    pFlowline.lIndex = lID
    pFlowline.iFlag_dam = 1 # this one is not dam, but it should be preserved
    aFlowline_out.append(pFlowline)
    pVertex_start = pFlowline.pVertex_start
    pVertex_end = pFlowline.pVertex_end
    lID = lID + 1    

    #we might find more than 1 upstream    
    def find_upstream_flowline(pVertex_start_in, pVertex_end_in):
        nUpstream = 0 
        aUpstream=list()
        aReverse=list()
        for i in range(nFlowline):
            pFlowline = aFlowline_in[i]
            pVerter_start = pFlowline.pVertex_start
            pVerter_end = pFlowline.pVertex_end
            if pVerter_end == pVertex_start_in  and pVerter_start!=pVertex_end_in:
                if aFlag_process[i] != 1:
                    nUpstream = nUpstream + 1
                    aUpstream.append(i)
                    aReverse.append(0)
                    aFlag_process[i] = 1
                    pass
            else:
                if pVerter_start == pVertex_start_in and pVerter_end !=pVertex_end_in :
                    if aFlag_process[i] != 1:
                        nUpstream = nUpstream + 1
                        aUpstream.append(i)
                        aReverse.append(1)
                        aFlag_process[i] = 1
                        pass
                pass

            pass
        return nUpstream, aUpstream, aReverse
    
    
    def tag_upstream(pVertex_start_in, pVertex_end_in):
        if(check_head_water(aFlowline_in, pVertex_start_in)==1):            
            pass
        else:
            nUpstream, aUpstream, aReverse = find_upstream_flowline(pVertex_start_in, pVertex_end_in)
            if nUpstream > 0:
                global lID
                for j in range(nUpstream):
                    pFlowline = aFlowline_in[ aUpstream[j] ] 
                    if (aReverse[j]==1):             
                        pFlowline.reverse()                    
                        pass
                    else:
                        pass
                    pFlowline.lIndex = lID
                    aFlowline_out.append(pFlowline)
                    lID = lID + 1
                    tag_upstream(  pFlowline.pVertex_start, pFlowline.pVertex_end  )            

                pass
            else:
                pass

    tag_upstream(pVertex_start, pVertex_end)
    
    return aFlowline_out


import os
from osgeo import osr
from pyflowline.formats.read_flowline import read_flowline_shapefile, read_flowline_shapefile_swat
from pyflowline.formats.export_flowline import export_flowline_to_json

#this one need to be updated
def convert_shapefile_to_json(iFlag_type_in, sFilename_shapefile_in, sFilename_geojson_in):
    """Convert a shapefile to a json file

    Args:
        iFlag_type_in (int): [0: vertex/point; 1: flowline/polyline; 2:polygon ]
        sFilename_shapefile_in ([type]): [description]
        sFilename_geojson_in ([type]): [description]
    """
    if os.path.isfile(sFilename_shapefile_in):
        pass
    else:
        print('This shapefile does not exist: ', sFilename_shapefile_in )
        iReturn_code = 0
        return iReturn_code


    if iFlag_type_in ==0:
        pass
    else:
        if iFlag_type_in == 1:
            aFlowline_basin, pSpatial_reference = read_flowline_shapefile( sFilename_shapefile_in )    
            #convert it

            iFlag_projected = 0
            pSpatial_reference_gcs = osr.SpatialReference()
            pSpatial_reference_gcs.ImportFromEPSG(4326)
            pSpatial_reference_gcs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
            export_flowline_to_json(aFlowline_basin, sFilename_geojson_in)
        else:
            if iFlag_type_in == 2:
                #polygon
                pass


    return

def convert_shapefile_to_json_swat(iFlag_type_in, sFilename_shapefile_in, sFilename_geojson_in):
    """Convert a shapefile to a json file

    Args:
        iFlag_type_in (int): [0: vertex/point; 1: flowline/polyline; 2:polygon ]
        sFilename_shapefile_in ([type]): [description]
        sFilename_geojson_in ([type]): [description]
    """

    if iFlag_type_in ==0:
        pass
    else:
        if iFlag_type_in == 1:
            aFlowline_basin, pSpatial_reference = read_flowline_shapefile_swat( sFilename_shapefile_in )     
            iFlag_projected = 0
            pSpatial_reference_gcs = osr.SpatialReference()
            pSpatial_reference_gcs.ImportFromEPSG(4326)
            pSpatial_reference_gcs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
            export_flowline_to_json(aFlowline_basin, sFilename_geojson_in)
        else:
            if iFlag_type_in == 2:
                #polygon
                pass


    return


import os, sys
from pathlib import Path
from os.path import realpath
#use this function to generate an initial json file for hexwatershed
import json
#once it's generated, you can modify it and use it for different simulations
from pyflowline.classes.pycase import flowlinecase
from pyflowline.classes.basin import pybasin

def pyflowline_generate_basin_template_configuration_file(sFilename_basins_json, nBasin, sWorkspace_input_in, sWorkspace_output_in):
    """generate baisn configureation

    Args:
        sFilename_basins_json (str): the filename
        nBasin (int): the total number of basin
        sWorkspace_input_in (str): the input data path
        sWorkspace_output (str): the output path

    Returns:
        basin: a basin object
    """

    aBasin_out = list()
    for i in range(nBasin):
        sBasin =  "{:03d}".format(i+1)   
        aConfig_basin = {}
        aConfig_basin['iFlag_dam'] = 0
        aConfig_basin['iFlag_disconnected'] = 0
        aConfig_basin['lBasinID'] = i + 1
        aConfig_basin['dLatitude_outlet_degree'] = -180
        aConfig_basin['dLongitude_outlet_degree'] = 180
        aConfig_basin['dAccumulation_threshold'] = -90
        aConfig_basin['dThreshold_small_river'] = 90        
        aConfig_basin['sFilename_dam'] = str(Path(sWorkspace_input_in)  /  'ICoM_dams.csv')        
        aConfig_basin['sFilename_flowline_filter'] = str(Path(sWorkspace_input_in)  /  'streamord7above.shp')        
        aConfig_basin['sFilename_flowline_raw'] = str(Path(sWorkspace_input_in)  /  'allflowline.shp')        
        aConfig_basin['sFilename_flowline_topo'] = str(Path(sWorkspace_input_in)  /  'flowline.csv')        
        aConfig_basin['sWorkspace_output_basin'] = str(Path(sWorkspace_output_in) / sBasin )
        pBasin = pybasin(aConfig_basin)    
        aBasin_out.append(pBasin)        
        pass
        
    #export basin config to a file    
    sFilename_basins_json = sFilename_basins_json 
    with open(sFilename_basins_json, 'w', encoding='utf-8') as f:
        sJson = json.dumps([json.loads(ob.tojson()) for ob in aBasin_out], indent = 4)        
        f.write(sJson)    
        f.close()

    return aBasin_out

def pyflowline_generate_template_configuration_file(sFilename_json, \
    sWorkspace_input, sWorkspace_output, iFlag_standalone_in=None, \
        iFlag_use_mesh_dem_in=None, iFlag_use_shapefile_extent_in=None,  iCase_index_in=None,  dResolution_degree_in = None, dResolution_meter_in = None,   sDate_in = None,  sMesh_type_in = None,    sModel_in = None):
    """generate pyflowline config template file

    Args:
        sFilename_json (str): _description_
        sWorkspace_input (str): _description_
        iFlag_standalone_in (int, optional): _description_. Defaults to None.
        iFlag_use_mesh_dem_in (int, optional): _description_. Defaults to None.
        iFlag_use_shapefile_extent_in (int, optional): _description_. Defaults to None.
        iCase_index_in (int, optional): _description_. Defaults to None.
        dResolution_degree_in (float, optional): _description_. Defaults to None.
        dResolution_meter_in (float, optional): _description_. Defaults to None.
        sDate_in (str, optional): _description_. Defaults to None.
        sMesh_type_in (str, optional): _description_. Defaults to None.
        sModel_in (str, optional): _description_. Defaults to None.
        sWorkspace_output_in (str, optional): _description_. Defaults to None.

    Returns:
        _type_: _description_
    """
    if os.path.exists(sFilename_json):         
        os.remove(sFilename_json)

    if iCase_index_in is not None:        
        iCase_index = iCase_index_in
    else:       
        iCase_index = 1
    
    if iFlag_standalone_in is not None:        
        iFlag_standalone = iFlag_standalone_in
    else:       
        iFlag_standalone = 1

    if iFlag_use_mesh_dem_in is not None:        
        iFlag_use_mesh_dem = iFlag_use_mesh_dem_in
    else:       
        iFlag_use_mesh_dem = 0

    if iFlag_use_shapefile_extent_in is not None:        
        iFlag_use_shapefile_extent = iFlag_use_shapefile_extent_in
    else:       
        iFlag_use_shapefile_extent = 0

    if sMesh_type_in is not None:
        sMesh_type = sMesh_type_in
    else:
        sMesh_type = 'hexagon'
        pass
    if sDate_in is not None:
        sDate = sDate_in
    else:
        sDate = '20220202'
        pass
    
    
    nBasin = 1

 
    #use a dict to initialize the class
    aConfig = {}
    
    
    aConfig['iFlag_use_shapefile_extent'] = iFlag_use_shapefile_extent 
    aConfig['iFlag_use_mesh_dem'] = iFlag_use_mesh_dem
    aConfig['iFlag_save_mesh'] = 1
    aConfig['iFlag_simplification']=1
    aConfig['iFlag_create_mesh']=1
    aConfig['iFlag_intersect']=1
    aConfig['iFlag_resample_method']=1
    aConfig['iFlag_global']=0
    aConfig['iFlag_multiple_outlet']=0
    aConfig['iFlag_elevation_profile']=1
    aConfig['iFlag_rotation']=0
    aConfig['iFlag_stream_burning_topology']=1
    aConfig['iFlag_save_elevation']=1     
    aConfig['nOutlet'] = nBasin
    aConfig['dResolution_degree'] = 0.5
    aConfig['dResolution_meter'] = 50000
    aConfig['dLongitude_left'] = -180
    aConfig['dLongitude_right'] = 180
    aConfig['dLatitude_bot'] = -90
    aConfig['dLatitude_top'] = 90
    aConfig['sFilename_model_configuration']  = sFilename_json 
    aConfig['sWorkspace_input'] = sWorkspace_input    
    aConfig['sWorkspace_output'] = sWorkspace_output        
    aConfig['sRegion'] = 'susquehanna'
    aConfig['sModel'] = 'pyflowline'
    aConfig['iCase_index'] = iCase_index   
    aConfig['sMesh_type'] = sMesh_type
    aConfig['sJob'] = 'pyflowline'
    aConfig['sDate']= sDate
    #full path
    aConfig['sFilename_mesh_netcdf'] = str(Path(sWorkspace_input)  /  'lnd_cull_mesh.nc')
    aConfig['sFilename_spatial_reference'] =  str(Path(sWorkspace_input)  /  'boundary_proj.shp')   
    #relative path
    aConfig['flowline_info'] = 'flowline_info.json'
    aConfig['sFilename_mesh_info'] = 'mesh_info.json'
    aConfig['sFilename_elevation'] = 'elevation.json' 
    oModel = flowlinecase(aConfig)
    #generate basin
    sDirname = os.path.dirname(sFilename_json)
    sFilename =  Path(sFilename_json).stem + '_basins.json'
    sFilename_basins_json = os.path.join(sDirname, sFilename)   
    aBasin = pyflowline_generate_basin_template_configuration_file(sFilename_basins_json, nBasin, sWorkspace_input,  oModel.sWorkspace_output)
    oModel.aBasin = aBasin
    oModel.sFilename_basins = sFilename_basins_json
    oModel.export_config_to_json(sFilename_json)    

    return oModel






from pyflowline.algorithms.auxiliary.find_index_in_list import find_flowline_in_list
def remove_duplicate_flowline(aFlowline_in):
    aFlowline_out = list()
    nFlowline = len(aFlowline_in)
    for i in range(nFlowline):
        pFlowline = aFlowline_in[i]
        iFlag_exist, lIndex = find_flowline_in_list( aFlowline_out,  pFlowline)
        if iFlag_exist ==1:
            pass
        else:
            aFlowline_out.append(pFlowline)

        pass
    return aFlowline_out



from abc import ABCMeta, abstractmethod
class mesh(object):
    """Abstract base class for
    """
    aParameter = {}
    dLatitude_bot = -90
    dLatitude_top = 90
    dLongitude_left = -180
    dLongitude_right = 180
    aRectangle=[]
    sFilename_mesh=''
    __metaclass__ = ABCMeta    
    @abstractmethod
    def __init__(self, aParameter):

        pass
    @abstractmethod
   
    def calculate_mesh_area(self):

        pass




from pyflowline.classes.vertex import pyvertex
from pyflowline.classes.edge import pyedge
from pyflowline.classes.flowline import pyflowline

from pyflowline.classes.hexagon import pyhexagon
from pyflowline.classes.square import pysquare
from pyflowline.classes.latlon import pylatlon
from pyflowline.classes.mpas import pympas
from pyflowline.classes.tin import pytin

from pyflowline.algorithms.auxiliary.gdal_functions import reproject_coordinates

def convert_gcs_coordinates_to_cell(iMesh_type_in, \
    dLongitude_center_in, \
    dLatitude_center_in, \
        aCoordinates_gcs_in):

    npoint = len(aCoordinates_gcs_in)    
    aVertex=list()              
    aEdge=list()    
    for i in range(npoint-1):
        x = aCoordinates_gcs_in[i][0]
        y = aCoordinates_gcs_in[i][1]
        dummy = dict()
        dummy['dLongitude_degree'] = x
        dummy['dLatitude_degree'] = y
        pVertex = pyvertex(dummy)
        aVertex.append(pVertex)

    npoint2 = len(aVertex) 
    for j in range(npoint2-1):
        pEdge = pyedge( aVertex[j], aVertex[j+1] )
        aEdge.append(pEdge)
    
    #add the last one    
    pEdge = pyedge( aVertex[npoint2-1], aVertex[0] )
    aEdge.append(pEdge)

    if iMesh_type_in ==1: #hexagon
        

        pHexagon = pyhexagon( dLongitude_center_in, dLatitude_center_in, aEdge, aVertex)
        return pHexagon
    else:
        if iMesh_type_in ==2: #sqaure
            pSquare = pysquare(dLongitude_center_in, dLatitude_center_in, aEdge, aVertex)
            return pSquare
        else:
            if iMesh_type_in ==3: #latlon
                pLatlon = pylatlon(dLongitude_center_in, dLatitude_center_in, aEdge, aVertex)
                return pLatlon
            else:
                if iMesh_type_in ==4: #mpas       
                    pMpas = pympas(  dLongitude_center_in, dLatitude_center_in, aEdge, aVertex)
                    return pMpas
                else:
                    if iMesh_type_in ==5: #tin
                        pTin = pytin(dLongitude_center_in, dLatitude_center_in, aEdge, aVertex)
                        return pTin
                        pass
                    else:
                        print('What mesh type are you using?')
                        return None

def convert_pcs_coordinates_to_cell(iMesh_type_in, aCoordinates_pcs_in, pSpatial_reference_in):

    npoint = len(aCoordinates_pcs_in)    
    aVertex=list()              
    aEdge=list()    


    for i in range(npoint):
        x = aCoordinates_pcs_in[i][0]
        y = aCoordinates_pcs_in[i][1]
        dummy = dict()
        dummy['x'] = x
        dummy['y'] = y

        dummy['dLongitude_degree'], dummy['dLatitude_degree'] = reproject_coordinates(x, y , pSpatial_reference_in)
        pVertex = pyvertex(dummy)
        aVertex.append(pVertex)
    for j in range(npoint-1):
        pEdge = pyedge( aVertex[j], aVertex[j+1] )
        aEdge.append(pEdge)

    

    if iMesh_type_in ==1: #hexagon     

        pHexagon = pyhexagon(  aEdge, aVertex)
        return pHexagon
    else:
        if iMesh_type_in ==2: #sqaure
            pSquare = pysquare( aEdge, aVertex)
            return pSquare
        else:
            if iMesh_type_in ==3: #latlon
                pLatlon = pylatlon( aEdge, aVertex)
                return pLatlon
            else:
                if iMesh_type_in ==4: #mpas   

                    pMpas = pympas( aEdge, aVertex)
                    return pMpas
                else:
                    if iMesh_type_in ==5: #tin
                        pTin = pytin( aEdge, aVertex)
                        return pTin
                        pass
                    else:
                        print('What mesh type are you using?')
                        return None


def convert_gcs_coordinates_to_flowline(aCoordinates_in):
    npoint = len(aCoordinates_in)
    
    aVertex=list()
    for i in range(npoint):
        x = aCoordinates_in[i][0]
        y = aCoordinates_in[i][1]
        dummy = dict()
        dummy['dLongitude_degree'] = x
        dummy['dLatitude_degree'] = y
        pVertex = pyvertex(dummy)
        aVertex.append(pVertex)
        
    aEdge=list()
    for j in range(npoint-1):
        pEdge = pyedge( aVertex[j], aVertex[j+1] )
        aEdge.append(pEdge)
    
    pLine = pyflowline( aEdge)    
    return pLine

def convert_pcs_coordinates_to_flowline(aCoordinates_in, pSpatial_reference_in):
    npoint = len(aCoordinates_in)
    
    aVertex=list()
    for i in range(npoint):
        x = aCoordinates_in[i][0]
        y = aCoordinates_in[i][1]
        dummy = dict()
        dummy['x'] =x
        dummy['y'] =y
        lon, lat = reproject_coordinates(x, y, pSpatial_reference_in)
        dummy['lon'] = lon
        dummy['lat'] = lat
        pVertex = pyvertex(dummy)
        aVertex.append(pVertex)
        
    aEdge=list()
    for j in range(npoint-1):
        pEdge = pyedge( aVertex[j], aVertex[j+1] )
        aEdge.append(pEdge)
    
    pLine = pyflowline( aEdge)
    
    return pLine


import os
from pathlib import Path
from abc import ABCMeta
import json
from json import JSONEncoder
import datetime
import numpy as np
from osgeo import ogr, osr
from pyflowline.classes.mpas import pympas
from pyflowline.classes.hexagon import pyhexagon
from pyflowline.classes.latlon import pylatlon
from pyflowline.classes.square import pysquare
from pyflowline.classes.vertex import pyvertex
from pyflowline.classes.basin import pybasin
from pyflowline.classes.flowline import pyflowline
from pyflowline.classes.edge import pyedge
from pyflowline.formats.read_mesh import read_mesh_json
from pyflowline.algorithms.auxiliary.gdal_functions import reproject_coordinates
from pyflowline.algorithms.auxiliary.gdal_functions  import degree_to_meter
from pyflowline.algorithms.auxiliary.gdal_functions  import meter_to_degree
from pyflowline.algorithms.auxiliary.gdal_functions import retrieve_geotiff_metadata
from pyflowline.mesh.hexagon.create_hexagon_mesh import create_hexagon_mesh
from pyflowline.mesh.latlon.create_latlon_mesh import create_latlon_mesh
from pyflowline.mesh.square.create_square_mesh import create_square_mesh
from pyflowline.mesh.mpas.create_mpas_mesh import create_mpas_mesh
from pyflowline.mesh.tin.create_tin_mesh import create_tin_mesh

pDate = datetime.datetime.today()
sDate_default = "{:04d}".format(pDate.year) + "{:02d}".format(pDate.month) + "{:02d}".format(pDate.day)

class CaseClassEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.float32):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, list):
            pass  
        if isinstance(obj, pyvertex):
            return json.loads(obj.tojson()) #lVertexID
        if isinstance(obj, pyedge):
            return obj.lEdgeID        
        if isinstance(obj, pyflowline):
            return obj.lFlowlineID
        if isinstance(obj, pyhexagon):
            return obj.lCellID
        if isinstance(obj, pysquare):
            return obj.lCellID
        if isinstance(obj, pylatlon):
            return obj.lCellID
        if isinstance(obj, pympas):
            return obj.lCellID        
        if isinstance(obj, pybasin):
            return obj.lBasinID    
            
        return JSONEncoder.default(self, obj)

class flowlinecase(object):
    __metaclass__ = ABCMeta
    iCase_index= 0
    sMesh_type = 1
    iFlag_standalone=1
    iFlag_flowline = 1
    iFlag_global = 0
    iFlag_multiple_outlet = 0
    iFlag_use_shapefile_extent=1
    iFlag_use_mesh_dem=0
    iFlag_save_mesh = 0
    iFlag_simplification = 1 #user can turn on/off
    iFlag_create_mesh=1
    iFlag_intersect = 1
    
    iFlag_rotation=0

    nOutlet = 1 #by default , there shoule ne only one ouelet
    dResolution_degree=0.0
    dResolution_meter=0.0
    dThreshold_small_river=0.0
    dLongitude_left = -180
    dLongitude_right = 180
    dLatitude_bot = -90
    dLatitude_top = 90
    sFilename_model_configuration=''
    sWorkspace_input=''      
    sWorkspace_output=''    
    #sWorkspace_output_case=''    
    sRegion=''
    sModel=''
    sMesh_type ='hexagon'

    sCase=''
    sDate=''    

    sFilename_spatial_reference=''
    sFilename_dem=''   

    sFilename_mesh=''
    sFilename_mesh_info=''
    sFilename_mesh_netcdf=''
    

    aBasin = list()
    aFlowline_simplified=list()
    aFlowline_conceptual=list()
    aCellID_outlet = list()
    aCell=list()

    
    def __init__(self, aConfig_in,\
            iFlag_standalone_in= None,\
            sModel_in = None,\
            sDate_in = None,\
            sWorkspace_output_in = None):
        #flags
        if iFlag_standalone_in is not None:
            self.iFlag_standalone = iFlag_standalone_in
        else:
            if 'iFlag_standalone' in aConfig_in:
                self.iFlag_standalone = int(aConfig_in['iFlag_standalone'])
            else:
                self.iFlag_standalone=1

  
        if 'iFlag_flowline' in aConfig_in:
            self.iFlag_flowline             = int(aConfig_in[ 'iFlag_flowline'])
        else:
            #without iFlag_flowline the model is a mesh generator
            self.iFlag_flowline=1
        
        if 'iFlag_global' in aConfig_in:
            self.iFlag_global             = int(aConfig_in[ 'iFlag_global'])
        
        if 'iFlag_multiple_outlet' in aConfig_in:
            self.iFlag_multiple_outlet             = int(aConfig_in[ 'iFlag_multiple_outlet'])  
        
        if 'iFlag_simplification' in aConfig_in:
            self.iFlag_simplification = int(aConfig_in['iFlag_simplification'])
            #if simplification is desired, then we must activate flowline flag
            #it can be turn off only when a previous simplification was done and you dont want to redo it
        else:
            self.iFlag_simplification  = 1
        if self.iFlag_simplification ==1:
            self.iFlag_flowline = 1

        if 'iFlag_create_mesh' in aConfig_in:
            self.iFlag_create_mesh = int(aConfig_in['iFlag_create_mesh'])  
        else:  
            self.iFlag_create_mesh=1
        
        if 'iFlag_save_mesh' in aConfig_in:
            self.iFlag_save_mesh             = int(aConfig_in[ 'iFlag_save_mesh'])
        
        if 'iFlag_use_mesh_dem' in aConfig_in:
            self.iFlag_use_mesh_dem = int(aConfig_in['iFlag_use_mesh_dem'])

        if 'iFlag_use_shapefile_extent' in aConfig_in:
            self.iFlag_use_shapefile_extent = int(aConfig_in['iFlag_use_shapefile_extent'])    

        if 'iFlag_rotation' in aConfig_in:
            self.iFlag_rotation = int(aConfig_in['iFlag_rotation'])
        

        if 'iFlag_intersect' in aConfig_in:
            self.iFlag_intersect = int(aConfig_in['iFlag_intersect'])
        else:
            self.iFlag_intersect=1

        if 'nOutlet' in aConfig_in:
            self.nOutlet = int(aConfig_in['nOutlet'])
             
        if 'iCase_index' in aConfig_in:
            iCase_index = int(aConfig_in['iCase_index'])
        else:
            iCase_index = 1
        sCase_index = "{:03d}".format( iCase_index )
        self.iCase_index =   iCase_index

        if 'dResolution_degree' in aConfig_in:
            self.dResolution_degree = float(aConfig_in['dResolution_degree']) 

        if 'dResolution_meter' in aConfig_in:
            self.dResolution_meter = float(aConfig_in['dResolution_meter']) 
        else:
            print('Please specify resolution.')

        if 'dLongitude_left' in aConfig_in:
            self.dLongitude_left = float(aConfig_in['dLongitude_left']) 

        if 'dLongitude_right' in aConfig_in:
            self.dLongitude_right = float(aConfig_in['dLongitude_right']) 

        if 'dLatitude_bot' in aConfig_in:
            self.dLatitude_bot = float(aConfig_in['dLatitude_bot']) 

        if 'dLatitude_top' in aConfig_in:
            self.dLatitude_top = float(aConfig_in['dLatitude_top']) 

        if 'sFilename_model_configuration' in aConfig_in:
            self.sFilename_model_configuration    = aConfig_in[ 'sFilename_model_configuration']

        if 'sFilename_spatial_reference' in aConfig_in:
            self.sFilename_spatial_reference = aConfig_in['sFilename_spatial_reference']

        if 'sFilename_dem' in aConfig_in:
            self.sFilename_dem = aConfig_in['sFilename_dem']

        if 'sFilename_mesh_netcdf' in aConfig_in:
            self.sFilename_mesh_netcdf = aConfig_in['sFilename_mesh_netcdf']
        
        if 'sWorkspace_bin' in aConfig_in:
            self.sWorkspace_bin= aConfig_in[ 'sWorkspace_bin']
            
        if 'sWorkspace_input' in aConfig_in:
            self.sWorkspace_input = aConfig_in[ 'sWorkspace_input']
        
       
        
        if sWorkspace_output_in is not None:
            self.sWorkspace_output = sWorkspace_output_in
        else:
            if 'sWorkspace_output' in aConfig_in:
                self.sWorkspace_output = aConfig_in[ 'sWorkspace_output']
        
        if 'sJob' in aConfig_in:
            self.sJob =  aConfig_in['sJob'] 
        if 'sRegion' in aConfig_in:
            self.sRegion               = aConfig_in[ 'sRegion']
        
        if sModel_in is not None:
            self.sModel = sModel_in
        else:
            if 'sModel' in aConfig_in:
                self.sModel                = aConfig_in[ 'sModel']

                      
        sDate   = aConfig_in[ 'sDate']
        if sDate is not None:
            self.sDate= sDate
        else:
            self.sDate = sDate_default
        
        
        sCase = self.sModel  + self.sDate + sCase_index
        self.sCase = sCase

        if 'sMesh_type' in aConfig_in:
            self.sMesh_type =  aConfig_in['sMesh_type']
        else:
            self.sMesh_type = 'hexagon'

                
        
        sMesh_type = self.sMesh_type
        if sMesh_type =='hexagon': #hexagon
            self.iMesh_type = 1
        else:
            if sMesh_type =='square': #square
                self.iMesh_type = 2
            else:
                if sMesh_type =='latlon': #latlon
                    self.iMesh_type = 3
                else:
                    if sMesh_type =='mpas': #mpas
                        self.iMesh_type = 4
                    else:
                        if sMesh_type =='tin': #tin
                            self.iMesh_type = 5
                        else:
                            print('Unsupported mesh type?')
        
        #the model can be run as part of hexwatershed or standalone
        if self.iFlag_standalone == 1:
            #in standalone case, will add case information and update output path
            sPath = str(Path(self.sWorkspace_output)  /  sCase)
            self.sWorkspace_output = sPath
        else:
            #use specified output path, also do not add output or input tag
            sPath = self.sWorkspace_output
        
        Path(sPath).mkdir(parents=True, exist_ok=True)

        self.aBasin = list()
    
        if self.iFlag_flowline==1:
            if 'sFilename_basins' in aConfig_in:
                self.sFilename_basins = aConfig_in['sFilename_basins']
                if os.path.isfile(self.sFilename_basins):
                    pass
                else:
                    print('This basin configuration file does not exist: ', self.sFilename_basins )
                    exit()
                with open(self.sFilename_basins) as json_file:
                    dummy_data = json.load(json_file)     
                    for i in range(self.nOutlet):
                        sBasin =  "{:03d}".format(i+1)   
                        dummy_basin = dummy_data[i]
                        dummy_basin['sWorkspace_output_basin'] = str(Path(self.sWorkspace_output) /     sBasin )

                        pBasin = pybasin(dummy_basin)

                        self.aBasin.append(pBasin)
            else:
                pass
        else:
            pass
     

        #model generated files         
        self.sFilename_mesh = os.path.join(str(Path(self.sWorkspace_output)  ) , sMesh_type + ".json" )               
        self.sFilename_mesh_info= os.path.join(str(Path(self.sWorkspace_output)  ) , sMesh_type + "_mesh_info.json"  )    
                
        return
        

    def convert_flowline_to_json(self):
        if self.iFlag_flowline == 1:
            for pBasin in self.aBasin:            
                pBasin.convert_flowline_to_json()
                pass

        return
 
    def flowline_simplification(self):
        aFlowline_out = list()   #store all the flowline
        #export the outlet into a single file
        aOutlet = list()
        if self.iFlag_simplification == 1: 
            for pBasin in self.aBasin:
                aFlowline_basin = pBasin.flowline_simplification()                
                aFlowline_out = aFlowline_out + aFlowline_basin
                aOutlet.append(pBasin.pVertex_outlet)
            
            self.aFlowline_simplified = aFlowline_out
              

        return aFlowline_out
    
    def mesh_generation(self):        
        if self.iFlag_create_mesh ==1:
            iFlag_global =  self.iFlag_global
            iMesh_type = self.iMesh_type
            iFlag_save_mesh = self.iFlag_save_mesh
            iFlag_rotation = self.iFlag_rotation
            dResolution_degree = self.dResolution_degree
            dResolution_meter = self.dResolution_meter
            sFilename_dem = self.sFilename_dem
            sFilename_spatial_reference = self.sFilename_spatial_reference
            sFilename_mesh = self.sFilename_mesh
            if iMesh_type !=4: #hexagon
                spatial_reference_target = osr.SpatialReference()  
                spatial_reference_target.ImportFromEPSG(4326)
                if self.iFlag_use_shapefile_extent==1:
                    pDriver_shapefile = ogr.GetDriverByName('Esri Shapefile')
                    pDataset_shapefile = pDriver_shapefile.Open(self.sFilename_spatial_reference, 0)
                    pLayer_shapefile = pDataset_shapefile.GetLayer(0)
                    pSpatial_reference = pLayer_shapefile.GetSpatialRef()
                    (dOriginX, dX_right, dY_bot, dOriginY) = pLayer_shapefile.GetExtent() # not in same order as ogrinfo
                    dLongitude_left,  dLatitude_bot= reproject_coordinates(dOriginX, dY_bot,pSpatial_reference,    spatial_reference_target)
                    dLongitude_right, dLatitude_top= reproject_coordinates(dX_right, dOriginY,pSpatial_reference,  spatial_reference_target)
                    dLatitude_mean = 0.5 * (dLatitude_top + dLatitude_bot)
                    pass
                else:
                    dPixelWidth, dOriginX, dOriginY, nrow, ncolumn, pSpatialRef_dem, pProjection, pGeotransform\
                         = retrieve_geotiff_metadata(sFilename_dem)

                    dY_bot = dOriginY - (nrow+1) * dPixelWidth
                    dLongitude_left,  dLatitude_bot= reproject_coordinates(dOriginX, dY_bot,pSpatialRef_dem,    spatial_reference_target)
                    dX_right = dOriginX + (ncolumn +1) * dPixelWidth

                    dLongitude_right, dLatitude_top= reproject_coordinates(dX_right, dOriginY,pSpatialRef_dem,  spatial_reference_target)
                    dLatitude_mean = 0.5 * (dLatitude_top + dLatitude_bot)
                    pass

                if dResolution_meter < 0:
                    #not used
                    pass
                else:
                    dResolution_degree = meter_to_degree(dResolution_meter, dLatitude_mean)

                dX_left = dOriginX
                dY_top = dOriginY
            else:
                pass

            if iMesh_type ==1: #hexagon
                #hexagon edge
                dResolution_meter = degree_to_meter(dLatitude_mean, dResolution_degree )
                dArea = np.power(dResolution_meter,2.0)
                dLength_edge = np.sqrt(  2.0 * dArea / (3.0* np.sqrt(3.0))  )
                if iFlag_rotation ==0:            
                    dX_spacing = dLength_edge * np.sqrt(3.0)
                    dY_spacing = dLength_edge * 1.5
                    ncolumn= int( (dX_right - dX_left) / dX_spacing )
                    nrow= int( (dY_top - dY_bot) / dY_spacing ) 
                else:            
                    dX_spacing = dLength_edge * 1.5
                    dY_spacing = dLength_edge * np.sqrt(3.0)    
                    ncolumn= int( (dX_right - dX_left) / dX_spacing )+1
                    nrow= int( (dY_top - dY_bot) / dY_spacing )

                aHexagon = create_hexagon_mesh(iFlag_rotation, dX_left, dY_bot, dResolution_meter, ncolumn, nrow, \
                    sFilename_mesh, sFilename_spatial_reference)
                return aHexagon
            else:
                if iMesh_type ==2: #sqaure
                    ncolumn= int( (dX_right - dX_left) / dResolution_meter )
                    nrow= int( (dY_top - dY_bot) / dResolution_meter )

                    aSquare = create_square_mesh(dX_left, dY_bot, dResolution_meter, ncolumn, nrow, \
                        sFilename_mesh, sFilename_spatial_reference)
                    return aSquare
                else:
                    if iMesh_type ==3: #latlon
                        dResolution_meter = degree_to_meter(dLatitude_mean, dResolution_degree)
                        dArea = np.power(dResolution_meter,2.0)
                        dLatitude_top    = self.dLatitude_top   
                        dLatitude_bot    = self.dLatitude_bot   
                        dLongitude_left  = self.dLongitude_left 
                        dLongitude_right = self.dLongitude_right
                        ncolumn= int( (dLongitude_right - dLongitude_left) / dResolution_degree )
                        nrow= int( (dLatitude_top - dLatitude_bot) / dResolution_degree )
                        aLatlon = create_latlon_mesh(dLongitude_left, dLatitude_bot, dResolution_degree, ncolumn, nrow, \
                            sFilename_mesh)
                        return aLatlon
                    else:
                        if iMesh_type == 4: #mpas
                            iFlag_use_mesh_dem = self.iFlag_use_mesh_dem
                            sFilename_mesh_netcdf = self.sFilename_mesh_netcdf
                            dLatitude_top    = self.dLatitude_top   
                            dLatitude_bot    = self.dLatitude_bot   
                            dLongitude_left  = self.dLongitude_left 
                            dLongitude_right = self.dLongitude_right
                            aMpas = create_mpas_mesh(iFlag_global, iFlag_use_mesh_dem, iFlag_save_mesh, \
                                  dLongitude_left, dLongitude_right,  dLatitude_top, dLatitude_bot, \
                                        sFilename_mesh_netcdf,      sFilename_mesh)
                            return aMpas
                        else:
                            if iMesh_type ==5: #tin this one need to be updated because central location issue
                                #tin edge
                                dArea = np.power(dResolution_meter,2.0)
                                dLength_edge = np.sqrt(  4.0 * dArea /  np.sqrt(3.0) )  
                                dX_shift = 0.5 * dLength_edge
                                dY_shift = 0.5 * dLength_edge * np.sqrt(3.0) 
                                dX_spacing = dX_shift * 2
                                dY_spacing = dY_shift
                                ncolumn= int( (dX_right - dX_left) / dX_shift )
                                nrow= int( (dY_top - dY_bot) / dY_spacing ) 
                                aTin = create_tin_mesh(dX_left, dY_bot, dResolution_meter, ncolumn, nrow,sFilename_mesh, sFilename_spatial_reference)
                                return aTin
                            else:
                                print('Unsupported mesh type?')
                                return
        else:
            pass
        return
    
    def reconstruct_topological_relationship(self, aCell_raw):
        iFlag_intersect = self.iFlag_intersect
        if iFlag_intersect == 1:
            iMesh_type = self.iMesh_type        
            sWorkspace_output = self.sWorkspace_output
            nOutlet = self.nOutlet
            sFilename_mesh=self.sFilename_mesh
            self.aCell, pSpatial_reference_mesh = read_mesh_json(iMesh_type, sFilename_mesh)
            self.aCell = self.merge_cell_info(aCell_raw)        
            aFlowline_conceptual = list()   #store all the flowline
            aCellID_outlet = list()
            aBasin = list()
            aCell_intersect=list()
        
            for pBasin in self.aBasin:
                aCell_intersect_basin = pBasin.reconstruct_topological_relationship(iMesh_type,sFilename_mesh)
                aFlowline_conceptual = aFlowline_conceptual + pBasin.aFlowline_basin_conceptual
                aBasin.append(pBasin)
                aCellID_outlet.append(pBasin.lCellID_outlet)
                aCell_intersect = aCell_intersect + aCell_intersect_basin

                #update length?
            for pCell in self.aCell:
                for pCell2 in aCell_intersect:
                    if pCell2.lCellID == pCell.lCellID:
                        pCell.dLength_flowline = pCell2.dLength_flowline
            
            self.aFlowline_conceptual = aFlowline_conceptual
            self.aCellID_outlet = aCellID_outlet

            return self.aCell, aFlowline_conceptual, aCellID_outlet
        else:
            return None
            

        
   
    def merge_cell_info(self, aCell_raw):

        for pCell in self.aCell:
            for pCell2 in aCell_raw:
                if pCell.lCellID == pCell2.lCellID:
                    pCell.aNeighbor = pCell2.aNeighbor
                    pCell.nNeighbor= pCell2.nNeighbor
                    pCell.aNeighbor_land= pCell2.aNeighbor_land
                    pCell.nNeighbor_land= pCell2.nNeighbor_land
                    pCell.aNeighbor_distance= pCell2.aNeighbor_distance

                    break

        return self.aCell
        
    def analyze(self):
        if self.iFlag_flowline == 1:
            for pBasin in self.aBasin:
                pBasin.analyze()
        return

    def setup(self):
        if self.iFlag_flowline == 1:
            self.convert_flowline_to_json()
        return

    def run(self):
        if self.iFlag_flowline == 1:
            self.flowline_simplification()        
            aCell = self.mesh_generation()
            aCell_out, a, b = self.reconstruct_topological_relationship(aCell)
        else:
            #only mesh generator
            aCell = self.mesh_generation()
            self.aCell = aCell      
            aCell_out = aCell
        
        return aCell_out

   
    def export(self):
        
        self.export_mesh_info_to_json()
        if self.iFlag_flowline ==1:
            for pBasin in self.aBasin:
                pBasin.export()

        self.tojson()

    def export_mesh_info_to_json(self):        
        aCell_all = self.aCell
        sFilename_json = self.sFilename_mesh_info
        ncell=len(aCell_all)
        iFlag_flowline = self.iFlag_flowline 
        if iFlag_flowline == 1: #if there is conceptual flowline 
            aCellID_outlet = self.aCellID_outlet
            aFlowline = self.aFlowline_conceptual
            nFlowline = len(aFlowline)
            for i in range(nFlowline):
                pFlowline = aFlowline[i]
                nEdge = pFlowline.nEdge
                nVertex = pFlowline.nVertex
                aEdge = pFlowline.aEdge
                iStream_segment = pFlowline.iStream_segment
                iStream_order = pFlowline.iStream_order
                for j in range(nEdge):
                    pEdge = aEdge[j]
                    pVertex_start = pEdge.pVertex_start
                    pVertex_end = pEdge.pVertex_end
                    for k in range(ncell):
                        pVertex_center = aCell_all[k].pVertex_center

                        if pVertex_center == pVertex_start:
                            aCell_all[k].iStream_segment_burned = iStream_segment
                            aCell_all[k].iStream_order_burned = iStream_order

                            for l in range(ncell):
                                pVertex_center2 = aCell_all[l].pVertex_center
                                lCellID = aCell_all[l].lCellID
                                if pVertex_center2 == pVertex_end:
                                    aCell_all[k].lCellID_downstream_burned = lCellID
                                    if lCellID in aCellID_outlet:
                                        aCell_all[l].iStream_segment_burned = iStream_segment
                                        aCell_all[l].iStream_order_burned = iStream_order

                                    break
                                
                                

                pass
        else:
            #only mesh, no flowline
            pass
        

        with open(sFilename_json, 'w', encoding='utf-8') as f:
            sJson = json.dumps([json.loads(ob.tojson()) for ob in aCell_all], indent = 4)        
            f.write(sJson)    
            f.close()

        return

    def tojson(self):  
        aSkip = ['aBasin', \
                'aFlowline_simplified','aFlowline_conceptual','aCellID_outlet',
                'aCell']      

        obj = self.__dict__.copy()
        for sKey in aSkip:
            obj.pop(sKey, None)
        sJson = json.dumps(obj,\
            sort_keys=True, \
                indent = 4, \
                    ensure_ascii=True, \
                        cls=CaseClassEncoder)
        return sJson

    def export_config_to_json(self, sFilename_output_in = None):

        if self.iFlag_standalone == 1:
            if sFilename_output_in is not None:
                sFilename_output = sFilename_output_in
            else:
                #use current output path
                sFilename_output = os.path.join(self.sWorkspace_output, 'configuration.json' )
            
            #all basins            
            sName = 'configuration_basin.json'
            sFilename_configuration  =  os.path.join( self.sWorkspace_output  , sName)
            with open(sFilename_configuration, 'w', encoding='utf-8') as f:
                sJson = json.dumps([json.loads(ob.tojson()) for ob in self.aBasin],\
                    sort_keys=True, \
                    indent = 4)   
                f.write(sJson)    
                f.close()
            #update
            self.sFilename_basins = sFilename_configuration
            
        else:
            if sFilename_output_in is not None:
                sFilename_output = sFilename_output_in
            else:
                #use parent path
                sPath = Path(self.sWorkspace_output)                
                sFilename_output = os.path.join(sPath.parent.absolute(), 'configuration.json' )
            #all basins
            sPath = Path(self.sWorkspace_output)
            sName = 'configuration_basin.json'
            sFilename_configuration  =  os.path.join( sPath.parent.absolute() , sName)
            with open(sFilename_configuration, 'w', encoding='utf-8') as f:
                sJson = json.dumps([json.loads(ob.tojson()) for ob in self.aBasin],\
                    sort_keys=True, \
                    indent = 4)   
                f.write(sJson)    
                f.close()
            #update for pyhexwatershed
            self.sFilename_basins = sFilename_configuration
            self.sWorkspace_output =   sPath.parent.absolute()


        aSkip = ['aBasin', \
                'aFlowline_simplified','aFlowline_conceptual','aCellID_outlet',
                'aCell']

        obj = self.__dict__.copy()
        for sKey in aSkip:
            obj.pop(sKey, None)

        with open(sFilename_output, 'w', encoding='utf-8') as f:
            json.dump(obj, f,sort_keys=True, \
                ensure_ascii=False, \
                indent=4, \
                cls=CaseClassEncoder)
        return


    def export_basin_config_to_json(self, sFilename_output_in= None):
        if self.iFlag_standalone == 1:
            if sFilename_output_in is not None:
                sFilename_output = sFilename_output_in
            else:
                #use current output path
                sName = 'configuration_basin.json'
                sFilename_output  =  os.path.join( self.sWorkspace_output  , sName)                
            
            #all basins 
            with open(sFilename_output, 'w', encoding='utf-8') as f:
                sJson = json.dumps([json.loads(ob.tojson()) for ob in self.aBasin],\
                    sort_keys=True, \
                    indent = 4)   
                f.write(sJson)    
                f.close()
            #update
            self.sFilename_basins = sFilename_output
            
        else:
            if sFilename_output_in is not None:
                sFilename_output = sFilename_output_in
            else:
                #use current output path
                sPath = Path(self.sWorkspace_output)
                sName = 'configuration_basin.json'
                sFilename_output  =  os.path.join( sPath.parent.absolute() , sName)
            
            #all basins                       
            with open(sFilename_output, 'w', encoding='utf-8') as f:
                sJson = json.dumps([json.loads(ob.tojson()) for ob in self.aBasin],\
                    sort_keys=True, \
                    indent = 4)   
                f.write(sJson)    
                f.close()
            #update for pyhexwatershed
            self.sFilename_basins = sFilename_output
            
        return


import abc
from abc import ABCMeta, abstractmethod

from pyflowline.classes.vertex import pyvertex
from pyflowline.classes.edge import pyedge

class pycell(metaclass=ABCMeta):
    
    @abstractmethod
    def __init__(self, aEdge):    
        pass

    @abstractmethod
    def calculate_cell_area(self, aEdge):
        pass




from abc import ABCMeta
import numpy as np
import json
from json import JSONEncoder
from pyflowline.algorithms.auxiliary.gdal_functions import calculate_distance_based_on_lon_lat

class VertexClassEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()        
        return JSONEncoder.default(self, obj)

class pyvertex(object):
    __metaclass__ = ABCMeta  
    lIndex=-1 #this index will be used for array
    lVertexID=-1
    lFlowlineID = -1  #we use this id only for intersect
    dX_meter=-9999
    dY_meter=-9999
    dZ_meter=-9999
    dLongitude_degree=0.0
    dLatitude_degree=0.0
    dLongitude_radian=0.0
    dLatitude_radian=0.0
    dElevation=0.0    
    def __init__(self, aParameter):
        if 'x' in aParameter:            
            self.dX_meter             = float(aParameter['x'])
        
        if 'y' in aParameter:            
            self.dY_meter             = float(aParameter['y'])
        
        if 'z' in aParameter:            
            self.dZ_meter             = float(aParameter['z'])
        
        #longitude and latitude are always required     
        try:     
            self.dLongitude_degree      = float(aParameter['dLongitude_degree'])                 
            self.dLatitude_degree       = float(aParameter['dLatitude_degree'])
        except:
            print('Initialization of vertex failed!')
        
        return
    
    def __eq__(self, other):
        iFlag = -1
        
        c = self.calculate_distance(other)
        if( c < 1.0E-6 ): #be careful
            iFlag = 1
        else:
            iFlag = 0       

        return iFlag

    def __ne__(self, other):
        return not self.__eq__(other)
    
    def calculate_distance(self, other):                
        lon1 = self.dLongitude_degree
        lat1 = self.dLatitude_degree    
        lon2 = other.dLongitude_degree
        lat2 = other.dLatitude_degree
        dDistance = calculate_distance_based_on_lon_lat(lon1,  lat1, lon2, lat2)        
        return dDistance
    
    def tojson(self):
        sJson = json.dumps(self.__dict__, \
                sort_keys=True, \
                indent = 4, \
                ensure_ascii=True, \
                cls=VertexClassEncoder)
        return sJson

  






import os
import numpy as np
from osgeo import ogr, osr
from shapely.wkt import loads

from pyflowline.formats.convert_coordinates import convert_gcs_coordinates_to_cell
from pyflowline.formats.convert_coordinates import convert_gcs_coordinates_to_flowline

def intersect_flowline_with_mesh(iMesh_type_in, sFilename_mesh_in, sFilename_flowline_in, sFilename_output_in):

    if  os.path.exists(sFilename_mesh_in) and  os.path.exists(sFilename_flowline_in) : 
        pass
    else:
        print('The input file does not exist')
        return

    if os.path.exists(sFilename_output_in): 
        os.remove(sFilename_output_in)

    
    pDriver_geojson = ogr.GetDriverByName( "GeoJSON")
    aCell=list()
    aCell_intersect=list()
   
    pDataset_mesh = pDriver_geojson.Open(sFilename_mesh_in, 0)
    pDataset_flowline = pDriver_geojson.Open(sFilename_flowline_in, 0)   

    pLayer_mesh = pDataset_mesh.GetLayer(0)
    pSpatial_reference_mesh = pLayer_mesh.GetSpatialRef()
    nfeature_mesh = pLayer_mesh.GetFeatureCount()

    pLayer_flowline = pDataset_flowline.GetLayer(0)
    pSpatial_reference_flowline = pLayer_flowline.GetSpatialRef()
    nfeature_flowline = pLayer_flowline.GetFeatureCount()
    pLayerDefinition = pLayer_flowline.GetLayerDefn()
    
    comparison = pSpatial_reference_mesh.IsSame(pSpatial_reference_flowline)
    if(comparison != 1):
        iFlag_transform = 1
        transform = osr.CoordinateTransformation(pSpatial_reference_mesh, pSpatial_reference_flowline)
    else:
        iFlag_transform = 0

    pDataset_out = pDriver_geojson.CreateDataSource(sFilename_output_in)

    pLayerOut = pDataset_out.CreateLayer('flowline', pSpatial_reference_flowline, ogr.wkbMultiLineString)
    # Add one attribute
    pLayerOut.CreateField(ogr.FieldDefn('id', ogr.OFTInteger64)) #long type for high resolution
    pLayerOut.CreateField(ogr.FieldDefn('iseg', ogr.OFTInteger)) #long type for high resolution
    pLayerOut.CreateField(ogr.FieldDefn('iord', ogr.OFTInteger)) #long type for high resolution
    pLayerDefn = pLayerOut.GetLayerDefn()
    pFeatureOut = ogr.Feature(pLayerDefn)       
    lID_flowline = 0          
    aFlowline_intersect_all=list()   
    for pFeature_mesh in pLayer_mesh:       
        pGeometry_mesh = pFeature_mesh.GetGeometryRef()        
        dummy0 = loads( pGeometry_mesh.ExportToWkt() )
        aCoords_gcs = dummy0.exterior.coords
        aCoords_gcs= np.array(aCoords_gcs)       

        lCellID = pFeature_mesh.GetField("id")
        dLon = pFeature_mesh.GetField("lon")
        dLat = pFeature_mesh.GetField("lat")        
        dArea = pFeature_mesh.GetField("area")
        if (iFlag_transform ==1): 
            pGeometry_mesh.Transform(transform)
        if (pGeometry_mesh.IsValid()):
            pass
        else:
            print('Geometry issue')

        pGeometrytype_mesh = pGeometry_mesh.GetGeometryName()
        if(pGeometrytype_mesh == 'POLYGON'):            
            pCell = convert_gcs_coordinates_to_cell(iMesh_type_in, dLon, dLat, aCoords_gcs)     
            pCell.lCellID = lCellID             
            pCell.dArea = dArea 
            pCell.dLength = pCell.calculate_edge_length()
            pCell.dLength_flowline = pCell.dLength                     
            aFlowline_intersect = list()
            iFlag_intersected = 0 
            for j in range (nfeature_flowline):
                pFeature_flowline = pLayer_flowline.GetFeature(j)
                pGeometry_flowline = pFeature_flowline.GetGeometryRef()
                iStream_segment = pFeature_flowline.GetField("iseg")
                iStream_order = pFeature_flowline.GetField("iord")
                if (pGeometry_flowline.IsValid()):
                    pass
                else:
                    print('Geometry issue')         

                iFlag_intersect = pGeometry_flowline.Intersects( pGeometry_mesh )
                if( iFlag_intersect == True):
                    iFlag_intersected = 1
                    pGeometry_intersect = pGeometry_flowline.Intersection(pGeometry_mesh)                     
                    pGeometrytype_intersect = pGeometry_intersect.GetGeometryName()
                    if pGeometrytype_intersect == 'LINESTRING':
                        pFeatureOut.SetGeometry(pGeometry_intersect)
                        pFeatureOut.SetField("id", lID_flowline)         
                        pFeatureOut.SetField("iseg", iStream_segment)    
                        pFeatureOut.SetField("iord", iStream_order)           
                        pLayerOut.CreateFeature(pFeatureOut)    
                        dummy = loads( pGeometry_intersect.ExportToWkt() )
                        aCoords = dummy.coords                
                        dummy1= np.array(aCoords)
                        pLine = convert_gcs_coordinates_to_flowline(dummy1)
                        pLine.calculate_length()
                        pLine.lIndex = lID_flowline
                        pLine.iStream_segment = iStream_segment
                        pLine.iStream_order = iStream_order
                        aFlowline_intersect.append(pLine)
                        aFlowline_intersect_all.append(pLine)
                        lID_flowline = lID_flowline + 1
                    
                    else:
                        if(pGeometrytype_intersect == 'MULTILINESTRING'):
                            aLine = ogr.ForceToLineString(pGeometry_intersect)
                            for Line in aLine: 
                                pFeatureOut.SetGeometry(Line)
                                pFeatureOut.SetField("id", lID_flowline)         
                                pFeatureOut.SetField("iseg", iStream_segment)    
                                pFeatureOut.SetField("iord", iStream_order)           
                                pLayerOut.CreateFeature(pFeatureOut)    
                                dummy = loads( Line.ExportToWkt() )
                                aCoords = dummy.coords
                                dummy1= np.array(aCoords)
                                pLine = convert_gcs_coordinates_to_flowline(dummy1)
                                pLine.calculate_length()
                                pLine.lIndex = lID_flowline
                                pLine.iStream_segment = iStream_segment
                                pLine.iStream_order = iStream_order
                                aFlowline_intersect.append(pLine)
                                aFlowline_intersect_all.append(pLine)
                                lID_flowline = lID_flowline + 1
                            pass
                        else:
                            pass                            

                else:
                    pass
     
            #now add back to the cell object
            pCell.aFlowline = aFlowline_intersect
            pCell.nFlowline = len(aFlowline_intersect)
            if iFlag_intersected ==1:     
                pCell.iFlag_intersected = 1 
                pCell.dLength_flowline = 0.0
                for i in range (pCell.nFlowline):
                    pFlowline = pCell.aFlowline[i]
                    dLength_flowline = pFlowline.dLength 
                    if ( dLength_flowline > pCell.dLength_flowline ):
                        pCell.dLength_flowline = dLength_flowline

                #replace flowline length if there is an actual flowline                    
                aCell_intersect.append(pCell)
                aCell.append(pCell)
            else:
                pCell.iFlag_intersected = 0   
                aCell.append(pCell)
                pass
            
        else:
            pass

    
    return  aCell,aCell_intersect, aFlowline_intersect_all


from abc import ABCMeta, abstractmethod
import json
from json import JSONEncoder
import numpy as np
from pyflowline.classes.vertex import pyvertex
from pyflowline.algorithms.auxiliary.gdal_functions import  calculate_angle_betwen_vertex, \
    calculate_polygon_area, calculate_distance_to_plane

class EdgeClassEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.float32):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, list):
            pass  
        if isinstance(obj, pyvertex):
            return json.loads(obj.tojson()) #lVertexID
          
            
        return JSONEncoder.default(self, obj)

class pyedge(object):
    __metaclass__ = ABCMeta 
    lEdgeID=-1
    lIndex=-1
    pVertex_start = None
    pVertex_end = None
    dLength=0.0    
    lIndex_upstream=-1
    lIndex_downstream=-1

    def __init__(self, pVertex_start_in, pVertex_end_in):
        if pVertex_start_in == pVertex_end_in:
            print('The two vertices are the same')    
        else:
            self.pVertex_start = pVertex_start_in
            self.pVertex_end = pVertex_end_in
            self.dLength = self.calculate_length()

        return

    def calculate_length(self):
        dLength =0.0
        dLength = self.pVertex_start.calculate_distance( self.pVertex_end)
        self.dLength= dLength
        return dLength

    def check_shared_vertex(self, other):
        '''
        check whether two edges are sharing the same vertex
        '''
        iFlag_shared =-1
        v0 = self.pVertex_start
        v1 = self.pVertex_end
        v2 = other.pVertex_start
        v3 = other.pVertex_end
        if v0 == v2 or v0 == v2 or v1==v2 or v1==v3:
            iFlag_shared =1
        else:
            iFlag_shared=0

        return iFlag_shared
    
    def check_upstream(self, other):
        iFlag_upstream =-1
        v0 = self.pVertex_start
        v1 = self.pVertex_end
        v2 = other.pVertex_start
        v3 = other.pVertex_end
        if v0 == v3:
            iFlag_upstream =1
        else:
            iFlag_upstream=0

        return iFlag_upstream

    def check_downstream(self, other):
        iFlag_downstream =-1
        v0 = self.pVertex_start
        v1 = self.pVertex_end
        v2 = other.pVertex_start
        v3 = other.pVertex_end
        if v1 == v2:
            iFlag_downstream =1
        else:
            iFlag_downstream=0

        return iFlag_downstream

    def reverse(self):
        v0 = self.pVertex_start
        v1 = self.pVertex_end
        self.pVertex_start = v1
        self.pVertex_end = v0
    
    def is_overlap(self, pEdge_in):
        iFlag_overlap = 0
        pVertex_start1 = self.pVertex_start
        pVertex_end1 = self.pVertex_end
        pVertex_start2 = pEdge_in.pVertex_start
        pVertex_end2 = pEdge_in.pVertex_end

        if pVertex_start1 == pVertex_start2 and pVertex_end1 == pVertex_end2:
            iFlag_overlap = 1
        else:
            if  pVertex_start1 == pVertex_end2 and pVertex_end1 == pVertex_start2:
                iFlag_overlap = 1
            else:
                iFlag_overlap = 0

        return iFlag_overlap

    def check_vertex_on_edge(self, pVertex_in):
        iFlag =0 
        dDistance = -1
        dDistance_plane = 9999
        pVertex_start = self.pVertex_start
        pVertex_end = self.pVertex_end
        self.dLength = pVertex_start.calculate_distance(pVertex_end)
        if pVertex_in != pVertex_start and pVertex_in!=pVertex_end:
            d1 = pVertex_start.calculate_distance(pVertex_in)            
            d2 = pVertex_end.calculate_distance(pVertex_in)  
            d3 = d1 + d2 - self.dLength
            angle3deg = calculate_angle_betwen_vertex(\
                 pVertex_start.dLongitude_degree, pVertex_start.dLatitude_degree,\
                 pVertex_in.dLongitude_degree, pVertex_in.dLatitude_degree,\
                 pVertex_end.dLongitude_degree,pVertex_end.dLatitude_degree)

            dDistance_plane = calculate_distance_to_plane(\
                 pVertex_start.dLongitude_degree, pVertex_start.dLatitude_degree,\
                 pVertex_in.dLongitude_degree, pVertex_in.dLatitude_degree,\
                 pVertex_end.dLongitude_degree,pVertex_end.dLatitude_degree)
            lons = [pVertex_start.dLongitude_degree,pVertex_in.dLongitude_degree,pVertex_end.dLongitude_degree]
            lats = [pVertex_start.dLatitude_degree, pVertex_in.dLatitude_degree, pVertex_end.dLatitude_degree]
            dArea = calculate_polygon_area(lons, lats)

            if  angle3deg > 178 and d3 < 1.0: #care
                iFlag = 1
                dDistance = d1
            else:
                iFlag = 0
            
        else:
                iFlag = 0 

        return iFlag, dDistance, dDistance_plane
    
    def __eq__(self, other):                
        iFlag_overlap = self.is_overlap(other)  
        return iFlag_overlap

    def __ne__(self, other):
        return not self.__eq__(other)

    def tojson(self):

        obj = self.__dict__.copy()
        
        sJson = json.dumps(obj, \
            sort_keys=True, \
                indent = 4, \
                    ensure_ascii=True, \
                        cls=EdgeClassEncoder)
        return sJson

